from antlr4 import *
from antlr4.error.ErrorListener import ErrorListener
from radb.RALexer import RALexer
from radb.RAParser import RAParser as RAParserBase
from radb.RAParserVisitor import RAParserVisitor

import radb.parse
from radb import utils

import logging
logger = logging.getLogger('ra')

######################################################################

def fixLiteralNames():
    """Go through the base (ANTLR-generated) RAParser and produce a new
    list of more meaningful literal token names.
    """
    names = list()
    for i, (literal, symbolic) in enumerate(zip(RAParserBase.literalNames,
                                                RAParserBase.symbolicNames)):
        if symbolic == '<INVALID>' or literal != '<INVALID>':
            # the literal name is already fine here:
            name = literal
        elif i in (RAParserBase.WS, RAParserBase.COMMENT,
                   RAParserBase.STRING, RAParserBase.NUMBER, RAParserBase.ID):
            # there is no canonoical literal representation for these
            # tokens, so use their symbolic names:
            name = '<{}>'.format(symbolic)
        else:
            # a canonoical literal representation can derived as
            # follows because of the conventions followed by
            # RALexer.g4:
            name = "'{}'".format(symbolic.lower().replace('_', ' '))
        names.append(name)
    return names

class RAParser(RAParserBase):
    """A subclass overriding the base (ANTLR-generated) RAParser's
    literalNames class attribute to provide more meaningful literal
    token names.

    When a token has multiple possible literal representations (e.g.,
    RAParser.NOT can be 'not' or 'NOT'), ANTLR4 assign '<INVALID>' as
    its literal name, which is very misleading in error-reporting.
    Thus, we fix RAParser.literalNames here (while we are at it, we
    also makes it possible to recover the canonical literal
    representation if there is one (e.g., RAParser.IS_NOT_NULL would
    be 'is not null').
    """
    literalNames = fixLiteralNames()

def literal(i):
    return RAParser.literalNames[i][1:-1] # to strip the surrounding quotes

def symbolic(i):
    return RAParser.symbolicNames[i]

def sqlstr_to_str(ss):
    return ss[1:-1].replace("''", "'")

def str_to_sqlstr(s):
    return "'" + s.replace("'", "''") + "'"

######################################################################

class ParsingError(Exception):
    pass

class RAErrorListener(ErrorListener):
    def __init__(self):
        super(RAErrorListener, self).__init__()

    def syntaxError(self, recognizer, offendingSymbol, line, column, msg, e):
        raise ParsingError('line {}:{} {}'.format(line, column, msg))

    def reportAmbiguity(self, recognizer, dfa, startIndex, stopIndex, exact, ambigAlts, configs):
        raise ParsingError('ambiguity: {}'.format(configs))

    def reportAttemptingFullContext(self, recognizer,
                                    dfa, startIndex, stopIndex, conflictingAlts, configs):
        raise ParsingError('attempting full context: {}'.format(configs))

    def reportContextSensitivity(self, recognizer,
                                 dfa, startIndex, stopIndex, prediction, configs):
        raise ParsingError('context sensitivity: {}'.format(configs))

######################################################################

class ASTBuilder(RAParserVisitor):

    def visitStringLiteralValExpr(self, ctx):
        return radb.ast.RAString(ctx.STRING().getText())

    def visitNumberLiteralValExpr(self, ctx):
        return radb.ast.RANumber(ctx.NUMBER().getText())

    def visitValExprParenthesized(self, ctx):
        return self.visit(ctx.valExpr())

    def visitFuncExpr(self, ctx):
        return radb.ast.FuncValExpr(ctx.ID().getText(),
                                    list() if ctx.listOfValExprs() is None else\
                                    self.visit(ctx.listOfValExprs()))

    def visitAttrRef(self, ctx):
        if ctx.getChildCount() == 1:
            return radb.ast.AttrRef(None, ctx.ID(0).getText())
        else:
            return radb.ast.AttrRef(ctx.ID(0).getText(), ctx.ID(1).getText())

    def visitValExprBinaryOp(self, ctx):
        left = self.visit(ctx.valExpr(0))
        op = ctx.getChild(1).symbol.type
        right = self.visit(ctx.valExpr(1))
        return radb.ast.ValExprBinaryOp(left, op, right)

    def visitValExprUnaryOp(self, ctx, op_idx=0):
        op = ctx.getChild(op_idx).symbol.type
        input = self.visit(ctx.valExpr())
        return radb.ast.ValExprUnaryOp(op, input)

    def visitMultDivExpr(self, ctx):
        return self.visitValExprBinaryOp(ctx)

    def visitPlusMinusExpr(self, ctx):
        return self.visitValExprBinaryOp(ctx)

    def visitConcatExpr(self, ctx):
        return self.visitValExprBinaryOp(ctx)

    def visitCompareExpr(self, ctx):
        return self.visitValExprBinaryOp(ctx)

    def visitLikeExpr(self, ctx):
        return self.visitValExprBinaryOp(ctx)

    def visitIsNullExpr(self, ctx):
        return self.visitValExprUnaryOp(ctx, op_idx=1)

    def visitIsNotNullExpr(self, ctx):
        return self.visitValExprUnaryOp(ctx, op_idx=1)

    def visitNotExpr(self, ctx):
        return self.visitValExprUnaryOp(ctx)

    def visitAndExpr(self, ctx):
        return self.visitValExprBinaryOp(ctx)

    def visitOrExpr(self, ctx):
        return self.visitValExprBinaryOp(ctx)

    def visitListOfValExprs(self, ctx):
        valExprs = list() if ctx.listOfValExprs() is None else self.visit(ctx.listOfValExprs())
        valExprs.insert(0, self.visit(ctx.valExpr()))
        return valExprs

    def visitListOfIDs(self, ctx):
        ids = list() if ctx.listOfIDs() is None else self.visit(ctx.listOfIDs())
        ids.insert(0, ctx.ID().getText())
        return ids

    def visitRelExprParenthesized(self, ctx):
        return self.visit(ctx.relExpr())

    def visitRelRef(self, ctx):
        return radb.ast.RelRef(ctx.ID().getText())

    def visitRenameExpr(self, ctx):
        return radb.ast.Rename(None if ctx.ID() is None else ctx.ID().getText(),
                               None if ctx.listOfIDs() is None else self.visit(ctx.listOfIDs()),
                               self.visit(ctx.relExpr()))

    def visitProjectExpr(self, ctx):
        return radb.ast.Project(self.visit(ctx.listOfValExprs()), self.visit(ctx.relExpr()))

    def visitSelectExpr(self, ctx):
        return radb.ast.Select(self.visit(ctx.valExpr()), self.visit(ctx.relExpr()))

    def visitCrossExpr(self, ctx):
        left = self.visit(ctx.relExpr(0))
        right = self.visit(ctx.relExpr(1))
        return radb.ast.Cross(left, right)

    def visitJoinExpr(self, ctx):
        left = self.visit(ctx.relExpr(0))
        cond = None if ctx.valExpr() is None else self.visit(ctx.valExpr())
        right = self.visit(ctx.relExpr(1))
        return radb.ast.Join(left, cond, right)

    def visitCrossExpr(self, ctx):
        left = self.visit(ctx.relExpr(0))
        right = self.visit(ctx.relExpr(1))
        return radb.ast.Cross(left, right)

    def visitUnionExpr(self, ctx):
        left = self.visit(ctx.relExpr(0))
        right = self.visit(ctx.relExpr(1))
        return radb.ast.Union(left, right)

    def visitDiffExpr(self, ctx):
        left = self.visit(ctx.relExpr(0))
        right = self.visit(ctx.relExpr(1))
        return radb.ast.Diff(left, right)

    def visitIntersectExpr(self, ctx):
        left = self.visit(ctx.relExpr(0))
        right = self.visit(ctx.relExpr(1))
        return radb.ast.Intersect(left, right)

    def visitDefinition(self, ctx):
        return radb.ast.Define(ctx.ID().getText(), self.visit(ctx.relExpr()))

    def visitListCommand(self, ctx):
        return radb.ast.Command(RAParser.LIST)

    def visitQuitCommand(self, ctx):
        return radb.ast.Command(RAParser.QUIT)

    def visitClearCommand(self, ctx):
        return radb.ast.CommandClear(view=(None if ctx.ID() is None else ctx.ID().getText()),
                                     force=(ctx.FORCE() is not None))

    def visitSaveCommand(self, ctx):
        return radb.ast.CommandSave(force=(ctx.FORCE() is not None),
                                    view=(None if ctx.ID() is None else ctx.ID().getText()),
                                    filename=(None if ctx.STRING() is None else\
                                              ctx.STRING().getText()))

    def visitSourceCommand(self, ctx):
        return radb.ast.CommandSource(ctx.STRING().getText())

    def visitSqlexecCommand(self, ctx):
        s = ctx.SQLEXEC_TEXT().getText().strip()
        s = utils.lreplace(s, literal(RAParser.ARG_L), '')
        s = utils.rreplace(s, literal(RAParser.ARG_R), '')
        return radb.ast.CommandSqlexec(s)

    def visitStatement(self, ctx):
        return self.visit(ctx.getChild(0))

    def visitProgram(self, ctx):
        return [ self.visit(statement) for statement in ctx.statement() ]

######################################################################

def one_statement_from_string(s):
    lexer = RALexer(InputStream(s))
    lexer._listeners = [RAErrorListener()]
    parser = RAParser(CommonTokenStream(lexer))
    parser._listeners = [RAErrorListener()]
    tree = parser.statement()
    # from antlr4.tree.Trees import Trees
    # print(Trees.toStringTree(tree, None, parser))
    ast = ASTBuilder().visit(tree)
    return ast

def statements_from_file(filename):
    lexer = RALexer(FileStream(filename))
    lexer._listeners = [RAErrorListener()]
    parser = RAParser(CommonTokenStream(lexer))
    parser._listeners = [RAErrorListener()]
    tree = parser.program()
    asts = ASTBuilder().visit(tree)
    return asts

######################################################################

def statement_state_transition(state, line, i):
    char = line[i]
    char2 = line[i:i+2]
    if state == 'normal' or state is None:
        if char == literal(RAParser.TERMINATOR):
            state = None
            if i+1 < len(line):
                line = line[0:i+1] # ignore rest of line
                logger.warning('rest of the line after "{}" (character {}) ignored'\
                               .format(literal(RAParser.TERMINATOR), i+1))
        elif char == "'":
            state = 'in_string'
        elif char2 == '//':
            i = len(line)-1 # effectively ignore rest of line
        elif char2 == '/*':
            state = 'in_comment'
            i += 1 # skip an extra char
        else:
            state = 'normal'
    elif state == 'in_string':
        if char2 == "''":
            i += 1 # skip an extra char
        elif char == "'":
            state = 'normal'
        else:
            pass
    elif state == 'in_comment':
        if char2 == '*/':
            state = 'normal'
            i += 1 # skip an extra char
        else:
            pass
    else:
        assert False
    return state, line, i+1

def is_input_buffer_empty(input_buffer):
    return all(line == '' or line.isspace() for line in input_buffer)

def statement_string_from_stdin(echo=False):
    prompts = { None: 'ra> ',
                'normal' : '..> ',
                'in_string' : '.s> ',
                'in_comment' : '.c> ' }
    states = set(prompts.keys())
    state = None
    input_buffer = list()
    try:
        while True:
            line = input(prompts[state])
            if echo:
                print(line)
            i = 0
            while i < len(line):
                state, line, i = statement_state_transition(state, line, i)
            input_buffer.append(line)
            if state is None:
                if not is_input_buffer_empty(input_buffer):
                    yield '\n'.join(input_buffer)
                input_buffer = list()
    except EOFError:
        if not is_input_buffer_empty(input_buffer):
            # treat as if a teminator has been entered
            input_buffer.append(literal(RAParser.TERMINATOR))
            yield '\n'.join(input_buffer);
        return

class RACompleter:
    def __init__(self):
        self.words = list()
        for i, w in enumerate(RAParser.literalNames):
            if w.startswith("'\\"):
                w = w.strip("'")
                if i in (RAParser.RENAME, RAParser.PROJECT, RAParser.SELECT, RAParser.JOIN):
                    self.words.append('{}{} {} '.format(w,
                                                        literal(RAParser.ARG_L),
                                                        literal(RAParser.ARG_R)))
                elif i in (RAParser.LIST, RAParser.QUIT):
                    self.words.append('{}{}'.format(w, literal(RAParser.TERMINATOR)))
                elif i == RAParser.SOURCE:
                    self.words.append("{} ''{}".format(w, literal(RAParser.TERMINATOR)))
                elif i == RAParser.SQLEXEC:
                    self.words.append('{}{} {}{}'.format(w,
                                                         literal(RAParser.ARG_L),
                                                         literal(RAParser.ARG_R),
                                                         literal(RAParser.TERMINATOR)))
                else:
                    self.words.append(w + ' ')
                if i == RAParser.JOIN: # join gets two completion options (theta + natural)
                    self.words.append(w + ' ')
        self.prefix = None
    def complete(self, prefix, index):
        if prefix != self.prefix:
            self.matching_words = [ w for w in self.words if w.startswith(prefix) ]
            self.prefix = prefix
        try:
            return self.matching_words[index]
        except IndexError:
            return None




from typing import List, Dict, Union, Optional
import copy
import re

from radb.utils import AutoNameEnum, CustomJSONSerializable

import logging
logger = logging.getLogger('ra')

class TypeSysError(Exception):
    pass

ValType = AutoNameEnum('ValType', (
    'BOOLEAN',
    'NUMBER',
    'STRING',
    'DATE',
    'DATETIME',
    'UNKNOWN',
    'ANY',
))

class FuncSpec:
    """Specification (i.e., signature) of a built-in function.  See
    :ref:`Operators and Functions` and :ref:`Type Checking` for an
    overview.

    """
    def __init__(self,
                 name: str, outtype: ValType,
                 required_argtypes: List[ValType],
                 optional_argtypes: List[ValType],
                 no_max_argc: bool=False):
        """Create a function specification.  If no_max_argc is True, the
        function can take an arbitrary number of additional arguments;
        in that case, optional_argtypes must not be empty, and its
        last element will be used as the type for additional
        arguments.
        """
        self.name = name
        self.outtype = outtype
        self.required_argtypes = required_argtypes
        self.optional_argtypes = optional_argtypes
        self.no_max_argc = no_max_argc
    def __str__(self):
        s_argtypes = list()
        for argtype in self.required_argtypes:
            s_argtypes.append(argtype.value)
        for i, argtype in enumerate(self.optional_argtypes):
            s_argtypes.append(argtype.value\
                              + ('*' if (i == len(self.optional_argtypes)-1 and self.no_max_argc)\
                                 else '?'))
        return '{}({}) -> {}'.format(self.name, ', '.join(s_argtypes), self.outtype.value)
    @staticmethod
    def from_config_line(line: str):
        """Construct a :class:`FuncSpec` object from a line.  See
        :ref:`Specification of Built-in Functions` for the
        specification format.

        """
        r_argtype = r'\w+[\?\*]?'
        r_argtypes = r'\s*(?:{r_argtype}(?:\s*,\s*{r_argtype})*)?\s*'.format(r_argtype=r_argtype)
        r_decl = r'\s*(?P<s_name>[^\s\(]*)\s*' +\
                 r'\((?P<s_argtypes>{r_argtypes})\)'.format(r_argtypes=r_argtypes) +\
                 r'\s*->\s*(?P<s_outtype>\w+)\s*(\#.*)?$'
        m = re.match(r_decl, line)
        if m is None:
            raise TypeSysError('cannot parse function declaration: {}'.format(line.strip()))
        name, s_argtypes, s_outtype = m.group('s_name', 's_argtypes', 's_outtype')
        outtype = ValType(s_outtype)
        required_argtypes = list()
        optional_argtypes = list()
        no_max_argc = False
        for s_argtype in re.findall(r_argtype, s_argtypes):
            argtype = ValType(s_argtype.rstrip('?*'))
            if s_argtype[-1] in ('?', '*'):
                if no_max_argc:
                    raise TypeSysError('only the very last argument can have *')
                if s_argtype[-1] == '*':
                    no_max_argc = True
                optional_argtypes.append(argtype)
            else:
                if len(optional_argtypes) > 0:
                    raise TypeSysError('required argument cannot follow an optional one')
                required_argtypes.append(argtype)
        return FuncSpec(name, outtype, required_argtypes, optional_argtypes, no_max_argc)

class ValTypeChecker:
    def __init__(self, default_decl_lines: str, decl_lines: str=None):
        default_decls = ValTypeChecker.decls_from_config_lines(default_decl_lines)
        if decl_lines is None:
            self.decls = default_decls
        else:
            self.decls = ValTypeChecker.decls_from_config_lines(decl_lines)
            for f in default_decls:
                if f in self.decls:
                    logger.info('default signature of "{}" overridden by configuration'.format(f))
                else:
                    self.decls[f] = default_decls[f]
    @staticmethod
    def decls_from_config_lines(lines):
        decls = dict()
        for i, line in enumerate(lines.split('\n')):
            line = line.rsplit('#', 1)[0].strip() # strip trailing comment
            if line == '':
                continue
            f = FuncSpec.from_config_line(line)
            if f.name not in decls:
                decls[f.name] = list()
            decls[f.name].append(f)
        return decls
    def can_be_used_as(self, this, other):
        if this == other:
            return True
        if other == ValType.ANY:
            return True
        if this == ValType.UNKNOWN:
            return True # relaxed semantics: allow UNKNOWN to be used as any type
        if this == ValType.STRING and other in (ValType.DATE, ValType.DATETIME):
            return True
        if this == ValType.DATE and other == ValType.DATETIME:
            return True
        return False
    def function_call(self, fname: str, argtypes: List[ValType]):
        if fname not in self.decls:
            logger.warning('function "{}" not recognized'.format(fname))
            return ValType.UNKNOWN
        for funcspec in self.decls[fname]:
            if len(argtypes) < len(funcspec.required_argtypes):
                continue
            if (not funcspec.no_max_argc)\
               and len(argtypes) > len(funcspec.required_argtypes)+len(funcspec.optional_argtypes):
                continue
            targets = copy.copy(funcspec.required_argtypes)
            for i in range(len(argtypes)-len(funcspec.required_argtypes)):
                targets.append(funcspec.optional_argtypes\
                               [min(i, len(funcspec.optional_argtypes)-1)])
            if all(self.can_be_used_as(argtype, target)\
                   for argtype, target in zip(argtypes, targets)):
                return funcspec.outtype
        raise TypeSysError('function "{}" cannot be applied to ({});'
                           ' correct signature(s):\n    {}'\
                           .format(fname, ', '.join(argtype.value for argtype in argtypes),
                                   '\n    '.join(str(funcspec) for funcspec in self.decls[fname])))

class AttrSpec(CustomJSONSerializable):
    """Specification of an attribute in a :class:`RelType`.  See
    :ref:`Relation Schema and Attribute References` for an overview.

    """
    def __init__(self, rel: Optional[str], name: Optional[str], type: ValType):
        """Construct a :class:`AttrSpec`, given the name of the relation where
        the attribute originally comes from (optional), the name of
        the attribute (optional), and the type of the attribute.

        """
        assert rel is None or isinstance(rel, str)
        assert name is None or isinstance(name, str)
        assert isinstance(type, ValType)
        self.rel = rel
        self.name = name
        self.type = type
    def __str__(self):
        return '{}.{}:{}'.format('_' if self.rel is None else self.rel,
                                 '_' if self.name is None else self.name,
                                 self.type.value)
    def str_ref_only(self):
        return '{}.{}'.format('_' if self.rel is None else self.rel,
                              '_' if self.name is None else self.name)
    def str_name_and_type_only(self):
        return '{}:{}'.format('_' if self.name is None else self.name,
                              self.type.value)
    def can_be_confused_with(self, other: 'AttrSpec'):
        """Check if this :class:`AttrSpec` can be potentially confused with
        another one, if both are presented as possibilities when
        resolving an :class:`.AttrRef`.

        """
        return (self.rel == other.rel or self.rel is None or other.rel is None) and\
            (self.name == other.name)

class RelType(CustomJSONSerializable):
    def __init__(self, tmp, attrs):
        assert isinstance(tmp, int)
        assert all(isinstance(attr, AttrSpec) for attr in attrs)
        self.tmp = tmp
        self.attrs = attrs
    def __str__(self):
        return '{}({})'.format(self.tmp, ', '.join(str(attr) for attr in self.attrs))
    def str_attr_names_and_types(self):
        return [attr.str_name_and_type_only() for attr in self.attrs]
    def sql_rel(self):
        return 'rat{}'.format(self.tmp)
    def sql_attr(self, i):
        return 'a{}'.format(i)
    def sql_attrs(self):
        return [self.sql_attr(i) for i in range(len(self.attrs))]



import copy

class ViewCollection:
    def __init__(self):
        self.raw_def_map = dict()
        self.bases_map = dict()

    def clone(self):
        other = ViewCollection()
        other.raw_def_map = copy.deepcopy(self.raw_def_map)
        other.bases_map = copy.deepcopy(self.bases_map)
        return other

    def list(self):
        return sorted(self.raw_def_map.keys())

    def raw_def(self, v):
        return self.raw_def_map.get(v, None)

    def find_dependents(self, v, recurse=False):
        assert v in self.raw_def_map
        dependents = set(u for u, bases in self.bases_map.items() if v in bases)
        if not recurse:
            return dependents
        more = dependents
        while len(more) > 0:
            new_more = set()
            for u in more:
                new_more |= (self.find_dependents(u) - dependents)
            more = new_more
            dependents |= more
        return dependents

    def topo(self, dst=None, seen=None):
        """List views in topological order (where dependent views come later).
        If dst is specified, list only those views that it depends on
        indirectly or directly, plus itself.  Assume that the
        dependency graph is acyclic (and will not verify that).

        This implementation is not particularly efficient, but it
        suffices for the size of the dependency graph we face.
        """
        assert dst is None or dst in self.raw_def_map
        if seen is None:
            seen = set()
        elif dst in seen:
            return list()
        # handle the case of dst=None by treating it as a dummy node
        # that depends on everything else.
        seen |= {dst}
        ancestors = [dst]
        for parent in (sorted(self.bases_map[dst]) if dst is not None else self.list()):
            for a in self.topo(parent, seen):
                assert a not in ancestors
                ancestors.insert(-1, a)
        if ancestors[-1] is None:
            return ancestors[0:-1]
        else:
            return ancestors

    def register(self, v, raw_def, bases):
        assert isinstance(bases, set)
        self.raw_def_map[v] = raw_def
        self.bases_map[v] = bases

    def clear(self, v=None):
        if v is None:
            self.raw_def_map = dict()
            self.bases_map = dict()
        else:
            for u in list(self.find_dependents(v, recurse=True)) + [v]:
                del self.bases_map[u]
                del self.raw_def_map[u]

if __name__ == '__main__':
    views = ViewCollection()
    views.register(1, None, {2, 3})
    views.register(2, None, {4})
    views.register(3, None, {4, 8})
    views.register(4, None, {5})
    views.register(5, None, {6, 7})
    views.register(6, None, set())
    views.register(7, None, {8})
    views.register(8, None, {6})
    print(views.topo(1))
    print(views.topo())



# Generated from RAParser.g4 by ANTLR 4.7
# encoding: utf-8
from antlr4 import *
from io import StringIO
from typing.io import TextIO
import sys

def serializedATN():
    with StringIO() as buf:
        buf.write("\3\u608b\ua72a\u8133\ub9ed\u417c\u3be7\u7786\u5964\3\64")
        buf.write("\u00b7\4\2\t\2\4\3\t\3\4\4\t\4\4\5\t\5\4\6\t\6\4\7\t\7")
        buf.write("\4\b\t\b\4\t\t\t\3\2\3\2\3\2\3\2\3\2\3\2\3\2\3\2\3\2\3")
        buf.write("\2\5\2\35\n\2\3\2\3\2\3\2\5\2\"\n\2\3\2\3\2\3\2\5\2\'")
        buf.write("\n\2\3\2\3\2\3\2\3\2\3\2\3\2\3\2\3\2\3\2\3\2\3\2\3\2\3")
        buf.write("\2\3\2\3\2\3\2\3\2\3\2\3\2\3\2\3\2\3\2\3\2\3\2\3\2\7\2")
        buf.write("B\n\2\f\2\16\2E\13\2\3\3\3\3\3\3\5\3J\n\3\3\4\3\4\3\4")
        buf.write("\5\4O\n\4\3\5\3\5\3\5\3\5\3\5\3\5\3\5\3\5\3\5\3\5\3\5")
        buf.write("\3\5\5\5]\n\5\3\5\5\5`\n\5\3\5\3\5\3\5\3\5\3\5\3\5\3\5")
        buf.write("\3\5\3\5\3\5\3\5\3\5\3\5\3\5\5\5p\n\5\3\5\3\5\3\5\3\5")
        buf.write("\3\5\3\5\5\5x\n\5\3\5\3\5\3\5\3\5\3\5\3\5\3\5\3\5\3\5")
        buf.write("\3\5\3\5\3\5\3\5\7\5\u0087\n\5\f\5\16\5\u008a\13\5\3\6")
        buf.write("\3\6\3\6\3\6\3\7\3\7\3\7\5\7\u0093\n\7\3\7\3\7\5\7\u0097")
        buf.write("\n\7\3\7\3\7\5\7\u009b\n\7\3\7\3\7\5\7\u009f\n\7\3\7\3")
        buf.write("\7\3\7\3\7\3\7\5\7\u00a6\n\7\3\b\3\b\3\b\5\b\u00ab\n\b")
        buf.write("\3\b\3\b\3\t\7\t\u00b0\n\t\f\t\16\t\u00b3\13\t\3\t\3\t")
        buf.write("\3\t\2\4\2\b\n\2\4\6\b\n\f\16\20\2\6\3\2\35\36\3\2\37")
        buf.write(" \3\2(-\4\2\22\22\35\35\2\u00d8\2&\3\2\2\2\4F\3\2\2\2")
        buf.write("\6K\3\2\2\2\bo\3\2\2\2\n\u008b\3\2\2\2\f\u00a5\3\2\2\2")
        buf.write("\16\u00aa\3\2\2\2\20\u00b1\3\2\2\2\22\23\b\2\1\2\23\'")
        buf.write("\7\20\2\2\24\'\7\21\2\2\25\26\7\"\2\2\26\27\5\2\2\2\27")
        buf.write("\30\7#\2\2\30\'\3\2\2\2\31\32\7\22\2\2\32\34\7\"\2\2\33")
        buf.write("\35\5\4\3\2\34\33\3\2\2\2\34\35\3\2\2\2\35\36\3\2\2\2")
        buf.write("\36\'\7#\2\2\37 \7\22\2\2 \"\7\33\2\2!\37\3\2\2\2!\"\3")
        buf.write("\2\2\2\"#\3\2\2\2#\'\7\22\2\2$%\7\17\2\2%\'\5\2\2\5&\22")
        buf.write("\3\2\2\2&\24\3\2\2\2&\25\3\2\2\2&\31\3\2\2\2&!\3\2\2\2")
        buf.write("&$\3\2\2\2\'C\3\2\2\2()\f\f\2\2)*\t\2\2\2*B\5\2\2\r+,")
        buf.write("\f\13\2\2,-\t\3\2\2-B\5\2\2\f./\f\n\2\2/\60\7!\2\2\60")
        buf.write("B\5\2\2\13\61\62\f\t\2\2\62\63\t\4\2\2\63B\5\2\2\n\64")
        buf.write("\65\f\b\2\2\65\66\7\f\2\2\66B\5\2\2\t\678\f\4\2\289\7")
        buf.write("\r\2\29B\5\2\2\5:;\f\3\2\2;<\7\16\2\2<B\5\2\2\4=>\f\7")
        buf.write("\2\2>B\7\b\2\2?@\f\6\2\2@B\7\t\2\2A(\3\2\2\2A+\3\2\2\2")
        buf.write("A.\3\2\2\2A\61\3\2\2\2A\64\3\2\2\2A\67\3\2\2\2A:\3\2\2")
        buf.write("\2A=\3\2\2\2A?\3\2\2\2BE\3\2\2\2CA\3\2\2\2CD\3\2\2\2D")
        buf.write("\3\3\2\2\2EC\3\2\2\2FI\5\2\2\2GH\7\34\2\2HJ\5\4\3\2IG")
        buf.write("\3\2\2\2IJ\3\2\2\2J\5\3\2\2\2KN\7\22\2\2LM\7\34\2\2MO")
        buf.write("\5\6\4\2NL\3\2\2\2NO\3\2\2\2O\7\3\2\2\2PQ\b\5\1\2QR\7")
        buf.write("\"\2\2RS\5\b\5\2ST\7#\2\2Tp\3\2\2\2Up\7\22\2\2VW\7\23")
        buf.write("\2\2W_\7$\2\2XY\7\22\2\2Y\\\7\'\2\2Z]\7\35\2\2[]\5\6\4")
        buf.write("\2\\Z\3\2\2\2\\[\3\2\2\2]`\3\2\2\2^`\5\6\4\2_X\3\2\2\2")
        buf.write("_^\3\2\2\2`a\3\2\2\2ab\7%\2\2bp\5\b\5\ncd\7\24\2\2de\7")
        buf.write("$\2\2ef\5\4\3\2fg\7%\2\2gh\5\b\5\thp\3\2\2\2ij\7\25\2")
        buf.write("\2jk\7$\2\2kl\5\2\2\2lm\7%\2\2mn\5\b\5\bnp\3\2\2\2oP\3")
        buf.write("\2\2\2oU\3\2\2\2oV\3\2\2\2oc\3\2\2\2oi\3\2\2\2p\u0088")
        buf.write("\3\2\2\2qr\f\7\2\2rw\7\26\2\2st\7$\2\2tu\5\2\2\2uv\7%")
        buf.write("\2\2vx\3\2\2\2ws\3\2\2\2wx\3\2\2\2xy\3\2\2\2y\u0087\5")
        buf.write("\b\5\bz{\f\6\2\2{|\7\27\2\2|\u0087\5\b\5\7}~\f\5\2\2~")
        buf.write("\177\7\30\2\2\177\u0087\5\b\5\6\u0080\u0081\f\4\2\2\u0081")
        buf.write("\u0082\7\31\2\2\u0082\u0087\5\b\5\5\u0083\u0084\f\3\2")
        buf.write("\2\u0084\u0085\7\32\2\2\u0085\u0087\5\b\5\4\u0086q\3\2")
        buf.write("\2\2\u0086z\3\2\2\2\u0086}\3\2\2\2\u0086\u0080\3\2\2\2")
        buf.write("\u0086\u0083\3\2\2\2\u0087\u008a\3\2\2\2\u0088\u0086\3")
        buf.write("\2\2\2\u0088\u0089\3\2\2\2\u0089\t\3\2\2\2\u008a\u0088")
        buf.write("\3\2\2\2\u008b\u008c\7\22\2\2\u008c\u008d\7&\2\2\u008d")
        buf.write("\u008e\5\b\5\2\u008e\13\3\2\2\2\u008f\u00a6\7.\2\2\u0090")
        buf.write("\u0096\7/\2\2\u0091\u0093\7\7\2\2\u0092\u0091\3\2\2\2")
        buf.write("\u0092\u0093\3\2\2\2\u0093\u0094\3\2\2\2\u0094\u0097\7")
        buf.write("\22\2\2\u0095\u0097\7\35\2\2\u0096\u0092\3\2\2\2\u0096")
        buf.write("\u0095\3\2\2\2\u0097\u00a6\3\2\2\2\u0098\u009a\7\60\2")
        buf.write("\2\u0099\u009b\7\7\2\2\u009a\u0099\3\2\2\2\u009a\u009b")
        buf.write("\3\2\2\2\u009b\u009c\3\2\2\2\u009c\u009e\t\5\2\2\u009d")
        buf.write("\u009f\7\20\2\2\u009e\u009d\3\2\2\2\u009e\u009f\3\2\2")
        buf.write("\2\u009f\u00a6\3\2\2\2\u00a0\u00a1\7\61\2\2\u00a1\u00a6")
        buf.write("\7\20\2\2\u00a2\u00a6\7\62\2\2\u00a3\u00a4\7\63\2\2\u00a4")
        buf.write("\u00a6\7\64\2\2\u00a5\u008f\3\2\2\2\u00a5\u0090\3\2\2")
        buf.write("\2\u00a5\u0098\3\2\2\2\u00a5\u00a0\3\2\2\2\u00a5\u00a2")
        buf.write("\3\2\2\2\u00a5\u00a3\3\2\2\2\u00a6\r\3\2\2\2\u00a7\u00ab")
        buf.write("\5\b\5\2\u00a8\u00ab\5\n\6\2\u00a9\u00ab\5\f\7\2\u00aa")
        buf.write("\u00a7\3\2\2\2\u00aa\u00a8\3\2\2\2\u00aa\u00a9\3\2\2\2")
        buf.write("\u00ab\u00ac\3\2\2\2\u00ac\u00ad\7\6\2\2\u00ad\17\3\2")
        buf.write("\2\2\u00ae\u00b0\5\16\b\2\u00af\u00ae\3\2\2\2\u00b0\u00b3")
        buf.write("\3\2\2\2\u00b1\u00af\3\2\2\2\u00b1\u00b2\3\2\2\2\u00b2")
        buf.write("\u00b4\3\2\2\2\u00b3\u00b1\3\2\2\2\u00b4\u00b5\7\2\2\3")
        buf.write("\u00b5\21\3\2\2\2\26\34!&ACIN\\_ow\u0086\u0088\u0092\u0096")
        buf.write("\u009a\u009e\u00a5\u00aa\u00b1")
        return buf.getvalue()


class RAParser ( Parser ):

    grammarFileName = "RAParser.g4"

    atn = ATNDeserializer().deserialize(serializedATN())

    decisionsToDFA = [ DFA(ds, i) for i, ds in enumerate(atn.decisionToState) ]

    sharedContextCache = PredictionContextCache()

    literalNames = [ "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>", 
                     "';'", "'!'", "<INVALID>", "<INVALID>", "<INVALID>", 
                     "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>", 
                     "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>", 
                     "'\\rename'", "'\\project'", "'\\select'", "'\\join'", 
                     "'\\cross'", "'\\union'", "'\\diff'", "'\\intersect'", 
                     "'.'", "','", "'*'", "'/'", "'+'", "'-'", "'||'", "'('", 
                     "')'", "'_{'", "'}'", "':-'", "':'", "'<='", "'<>'", 
                     "'>='", "'<'", "'='", "'>'", "'\\list'", "'\\clear'", 
                     "'\\save'", "'\\source'", "'\\quit'", "'\\sqlexec'" ]

    symbolicNames = [ "<INVALID>", "WS", "COMMENT", "LINE_COMMENT", "TERMINATOR", 
                      "FORCE", "IS_NULL", "IS_NOT_NULL", "IS", "NULL", "LIKE", 
                      "AND", "OR", "NOT", "STRING", "NUMBER", "ID", "RENAME", 
                      "PROJECT", "SELECT", "JOIN", "CROSS", "UNION", "DIFF", 
                      "INTERSECT", "DOT", "COMMA", "STAR", "SLASH", "PLUS", 
                      "MINUS", "CONCAT", "PAREN_L", "PAREN_R", "ARG_L", 
                      "ARG_R", "GETS", "COLON", "LE", "NE", "GE", "LT", 
                      "EQ", "GT", "LIST", "CLEAR", "SAVE", "SOURCE", "QUIT", 
                      "SQLEXEC", "SQLEXEC_TEXT" ]

    RULE_valExpr = 0
    RULE_listOfValExprs = 1
    RULE_listOfIDs = 2
    RULE_relExpr = 3
    RULE_definition = 4
    RULE_command = 5
    RULE_statement = 6
    RULE_program = 7

    ruleNames =  [ "valExpr", "listOfValExprs", "listOfIDs", "relExpr", 
                   "definition", "command", "statement", "program" ]

    EOF = Token.EOF
    WS=1
    COMMENT=2
    LINE_COMMENT=3
    TERMINATOR=4
    FORCE=5
    IS_NULL=6
    IS_NOT_NULL=7
    IS=8
    NULL=9
    LIKE=10
    AND=11
    OR=12
    NOT=13
    STRING=14
    NUMBER=15
    ID=16
    RENAME=17
    PROJECT=18
    SELECT=19
    JOIN=20
    CROSS=21
    UNION=22
    DIFF=23
    INTERSECT=24
    DOT=25
    COMMA=26
    STAR=27
    SLASH=28
    PLUS=29
    MINUS=30
    CONCAT=31
    PAREN_L=32
    PAREN_R=33
    ARG_L=34
    ARG_R=35
    GETS=36
    COLON=37
    LE=38
    NE=39
    GE=40
    LT=41
    EQ=42
    GT=43
    LIST=44
    CLEAR=45
    SAVE=46
    SOURCE=47
    QUIT=48
    SQLEXEC=49
    SQLEXEC_TEXT=50

    def __init__(self, input:TokenStream, output:TextIO = sys.stdout):
        super().__init__(input, output)
        self.checkVersion("4.7")
        self._interp = ParserATNSimulator(self, self.atn, self.decisionsToDFA, self.sharedContextCache)
        self._predicates = None



    class ValExprContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser


        def getRuleIndex(self):
            return RAParser.RULE_valExpr

     
        def copyFrom(self, ctx:ParserRuleContext):
            super().copyFrom(ctx)


    class AndExprContext(ValExprContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a RAParser.ValExprContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def valExpr(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(RAParser.ValExprContext)
            else:
                return self.getTypedRuleContext(RAParser.ValExprContext,i)

        def AND(self):
            return self.getToken(RAParser.AND, 0)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitAndExpr" ):
                return visitor.visitAndExpr(self)
            else:
                return visitor.visitChildren(self)


    class NumberLiteralValExprContext(ValExprContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a RAParser.ValExprContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def NUMBER(self):
            return self.getToken(RAParser.NUMBER, 0)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitNumberLiteralValExpr" ):
                return visitor.visitNumberLiteralValExpr(self)
            else:
                return visitor.visitChildren(self)


    class LikeExprContext(ValExprContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a RAParser.ValExprContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def valExpr(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(RAParser.ValExprContext)
            else:
                return self.getTypedRuleContext(RAParser.ValExprContext,i)

        def LIKE(self):
            return self.getToken(RAParser.LIKE, 0)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitLikeExpr" ):
                return visitor.visitLikeExpr(self)
            else:
                return visitor.visitChildren(self)


    class FuncExprContext(ValExprContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a RAParser.ValExprContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def ID(self):
            return self.getToken(RAParser.ID, 0)
        def PAREN_L(self):
            return self.getToken(RAParser.PAREN_L, 0)
        def PAREN_R(self):
            return self.getToken(RAParser.PAREN_R, 0)
        def listOfValExprs(self):
            return self.getTypedRuleContext(RAParser.ListOfValExprsContext,0)


        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitFuncExpr" ):
                return visitor.visitFuncExpr(self)
            else:
                return visitor.visitChildren(self)


    class PlusMinusExprContext(ValExprContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a RAParser.ValExprContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def valExpr(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(RAParser.ValExprContext)
            else:
                return self.getTypedRuleContext(RAParser.ValExprContext,i)

        def PLUS(self):
            return self.getToken(RAParser.PLUS, 0)
        def MINUS(self):
            return self.getToken(RAParser.MINUS, 0)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitPlusMinusExpr" ):
                return visitor.visitPlusMinusExpr(self)
            else:
                return visitor.visitChildren(self)


    class CompareExprContext(ValExprContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a RAParser.ValExprContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def valExpr(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(RAParser.ValExprContext)
            else:
                return self.getTypedRuleContext(RAParser.ValExprContext,i)

        def LT(self):
            return self.getToken(RAParser.LT, 0)
        def LE(self):
            return self.getToken(RAParser.LE, 0)
        def EQ(self):
            return self.getToken(RAParser.EQ, 0)
        def NE(self):
            return self.getToken(RAParser.NE, 0)
        def GE(self):
            return self.getToken(RAParser.GE, 0)
        def GT(self):
            return self.getToken(RAParser.GT, 0)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitCompareExpr" ):
                return visitor.visitCompareExpr(self)
            else:
                return visitor.visitChildren(self)


    class OrExprContext(ValExprContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a RAParser.ValExprContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def valExpr(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(RAParser.ValExprContext)
            else:
                return self.getTypedRuleContext(RAParser.ValExprContext,i)

        def OR(self):
            return self.getToken(RAParser.OR, 0)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitOrExpr" ):
                return visitor.visitOrExpr(self)
            else:
                return visitor.visitChildren(self)


    class AttrRefContext(ValExprContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a RAParser.ValExprContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def ID(self, i:int=None):
            if i is None:
                return self.getTokens(RAParser.ID)
            else:
                return self.getToken(RAParser.ID, i)
        def DOT(self):
            return self.getToken(RAParser.DOT, 0)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitAttrRef" ):
                return visitor.visitAttrRef(self)
            else:
                return visitor.visitChildren(self)


    class MultDivExprContext(ValExprContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a RAParser.ValExprContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def valExpr(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(RAParser.ValExprContext)
            else:
                return self.getTypedRuleContext(RAParser.ValExprContext,i)

        def STAR(self):
            return self.getToken(RAParser.STAR, 0)
        def SLASH(self):
            return self.getToken(RAParser.SLASH, 0)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitMultDivExpr" ):
                return visitor.visitMultDivExpr(self)
            else:
                return visitor.visitChildren(self)


    class ConcatExprContext(ValExprContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a RAParser.ValExprContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def valExpr(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(RAParser.ValExprContext)
            else:
                return self.getTypedRuleContext(RAParser.ValExprContext,i)

        def CONCAT(self):
            return self.getToken(RAParser.CONCAT, 0)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitConcatExpr" ):
                return visitor.visitConcatExpr(self)
            else:
                return visitor.visitChildren(self)


    class ValExprParenthesizedContext(ValExprContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a RAParser.ValExprContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def PAREN_L(self):
            return self.getToken(RAParser.PAREN_L, 0)
        def valExpr(self):
            return self.getTypedRuleContext(RAParser.ValExprContext,0)

        def PAREN_R(self):
            return self.getToken(RAParser.PAREN_R, 0)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitValExprParenthesized" ):
                return visitor.visitValExprParenthesized(self)
            else:
                return visitor.visitChildren(self)


    class IsNullExprContext(ValExprContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a RAParser.ValExprContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def valExpr(self):
            return self.getTypedRuleContext(RAParser.ValExprContext,0)

        def IS_NULL(self):
            return self.getToken(RAParser.IS_NULL, 0)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitIsNullExpr" ):
                return visitor.visitIsNullExpr(self)
            else:
                return visitor.visitChildren(self)


    class IsNotNullExprContext(ValExprContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a RAParser.ValExprContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def valExpr(self):
            return self.getTypedRuleContext(RAParser.ValExprContext,0)

        def IS_NOT_NULL(self):
            return self.getToken(RAParser.IS_NOT_NULL, 0)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitIsNotNullExpr" ):
                return visitor.visitIsNotNullExpr(self)
            else:
                return visitor.visitChildren(self)


    class NotExprContext(ValExprContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a RAParser.ValExprContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def NOT(self):
            return self.getToken(RAParser.NOT, 0)
        def valExpr(self):
            return self.getTypedRuleContext(RAParser.ValExprContext,0)


        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitNotExpr" ):
                return visitor.visitNotExpr(self)
            else:
                return visitor.visitChildren(self)


    class StringLiteralValExprContext(ValExprContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a RAParser.ValExprContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def STRING(self):
            return self.getToken(RAParser.STRING, 0)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitStringLiteralValExpr" ):
                return visitor.visitStringLiteralValExpr(self)
            else:
                return visitor.visitChildren(self)



    def valExpr(self, _p:int=0):
        _parentctx = self._ctx
        _parentState = self.state
        localctx = RAParser.ValExprContext(self, self._ctx, _parentState)
        _prevctx = localctx
        _startState = 0
        self.enterRecursionRule(localctx, 0, self.RULE_valExpr, _p)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 36
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input,2,self._ctx)
            if la_ == 1:
                localctx = RAParser.StringLiteralValExprContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx

                self.state = 17
                self.match(RAParser.STRING)
                pass

            elif la_ == 2:
                localctx = RAParser.NumberLiteralValExprContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx
                self.state = 18
                self.match(RAParser.NUMBER)
                pass

            elif la_ == 3:
                localctx = RAParser.ValExprParenthesizedContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx
                self.state = 19
                self.match(RAParser.PAREN_L)
                self.state = 20
                self.valExpr(0)
                self.state = 21
                self.match(RAParser.PAREN_R)
                pass

            elif la_ == 4:
                localctx = RAParser.FuncExprContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx
                self.state = 23
                self.match(RAParser.ID)
                self.state = 24
                self.match(RAParser.PAREN_L)
                self.state = 26
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << RAParser.NOT) | (1 << RAParser.STRING) | (1 << RAParser.NUMBER) | (1 << RAParser.ID) | (1 << RAParser.PAREN_L))) != 0):
                    self.state = 25
                    self.listOfValExprs()


                self.state = 28
                self.match(RAParser.PAREN_R)
                pass

            elif la_ == 5:
                localctx = RAParser.AttrRefContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx
                self.state = 31
                self._errHandler.sync(self)
                la_ = self._interp.adaptivePredict(self._input,1,self._ctx)
                if la_ == 1:
                    self.state = 29
                    self.match(RAParser.ID)
                    self.state = 30
                    self.match(RAParser.DOT)


                self.state = 33
                self.match(RAParser.ID)
                pass

            elif la_ == 6:
                localctx = RAParser.NotExprContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx
                self.state = 34
                self.match(RAParser.NOT)
                self.state = 35
                self.valExpr(3)
                pass


            self._ctx.stop = self._input.LT(-1)
            self.state = 65
            self._errHandler.sync(self)
            _alt = self._interp.adaptivePredict(self._input,4,self._ctx)
            while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:
                if _alt==1:
                    if self._parseListeners is not None:
                        self.triggerExitRuleEvent()
                    _prevctx = localctx
                    self.state = 63
                    self._errHandler.sync(self)
                    la_ = self._interp.adaptivePredict(self._input,3,self._ctx)
                    if la_ == 1:
                        localctx = RAParser.MultDivExprContext(self, RAParser.ValExprContext(self, _parentctx, _parentState))
                        self.pushNewRecursionContext(localctx, _startState, self.RULE_valExpr)
                        self.state = 38
                        if not self.precpred(self._ctx, 10):
                            from antlr4.error.Errors import FailedPredicateException
                            raise FailedPredicateException(self, "self.precpred(self._ctx, 10)")
                        self.state = 39
                        _la = self._input.LA(1)
                        if not(_la==RAParser.STAR or _la==RAParser.SLASH):
                            self._errHandler.recoverInline(self)
                        else:
                            self._errHandler.reportMatch(self)
                            self.consume()
                        self.state = 40
                        self.valExpr(11)
                        pass

                    elif la_ == 2:
                        localctx = RAParser.PlusMinusExprContext(self, RAParser.ValExprContext(self, _parentctx, _parentState))
                        self.pushNewRecursionContext(localctx, _startState, self.RULE_valExpr)
                        self.state = 41
                        if not self.precpred(self._ctx, 9):
                            from antlr4.error.Errors import FailedPredicateException
                            raise FailedPredicateException(self, "self.precpred(self._ctx, 9)")
                        self.state = 42
                        _la = self._input.LA(1)
                        if not(_la==RAParser.PLUS or _la==RAParser.MINUS):
                            self._errHandler.recoverInline(self)
                        else:
                            self._errHandler.reportMatch(self)
                            self.consume()
                        self.state = 43
                        self.valExpr(10)
                        pass

                    elif la_ == 3:
                        localctx = RAParser.ConcatExprContext(self, RAParser.ValExprContext(self, _parentctx, _parentState))
                        self.pushNewRecursionContext(localctx, _startState, self.RULE_valExpr)
                        self.state = 44
                        if not self.precpred(self._ctx, 8):
                            from antlr4.error.Errors import FailedPredicateException
                            raise FailedPredicateException(self, "self.precpred(self._ctx, 8)")
                        self.state = 45
                        self.match(RAParser.CONCAT)
                        self.state = 46
                        self.valExpr(9)
                        pass

                    elif la_ == 4:
                        localctx = RAParser.CompareExprContext(self, RAParser.ValExprContext(self, _parentctx, _parentState))
                        self.pushNewRecursionContext(localctx, _startState, self.RULE_valExpr)
                        self.state = 47
                        if not self.precpred(self._ctx, 7):
                            from antlr4.error.Errors import FailedPredicateException
                            raise FailedPredicateException(self, "self.precpred(self._ctx, 7)")
                        self.state = 48
                        _la = self._input.LA(1)
                        if not((((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << RAParser.LE) | (1 << RAParser.NE) | (1 << RAParser.GE) | (1 << RAParser.LT) | (1 << RAParser.EQ) | (1 << RAParser.GT))) != 0)):
                            self._errHandler.recoverInline(self)
                        else:
                            self._errHandler.reportMatch(self)
                            self.consume()
                        self.state = 49
                        self.valExpr(8)
                        pass

                    elif la_ == 5:
                        localctx = RAParser.LikeExprContext(self, RAParser.ValExprContext(self, _parentctx, _parentState))
                        self.pushNewRecursionContext(localctx, _startState, self.RULE_valExpr)
                        self.state = 50
                        if not self.precpred(self._ctx, 6):
                            from antlr4.error.Errors import FailedPredicateException
                            raise FailedPredicateException(self, "self.precpred(self._ctx, 6)")
                        self.state = 51
                        self.match(RAParser.LIKE)
                        self.state = 52
                        self.valExpr(7)
                        pass

                    elif la_ == 6:
                        localctx = RAParser.AndExprContext(self, RAParser.ValExprContext(self, _parentctx, _parentState))
                        self.pushNewRecursionContext(localctx, _startState, self.RULE_valExpr)
                        self.state = 53
                        if not self.precpred(self._ctx, 2):
                            from antlr4.error.Errors import FailedPredicateException
                            raise FailedPredicateException(self, "self.precpred(self._ctx, 2)")
                        self.state = 54
                        self.match(RAParser.AND)
                        self.state = 55
                        self.valExpr(3)
                        pass

                    elif la_ == 7:
                        localctx = RAParser.OrExprContext(self, RAParser.ValExprContext(self, _parentctx, _parentState))
                        self.pushNewRecursionContext(localctx, _startState, self.RULE_valExpr)
                        self.state = 56
                        if not self.precpred(self._ctx, 1):
                            from antlr4.error.Errors import FailedPredicateException
                            raise FailedPredicateException(self, "self.precpred(self._ctx, 1)")
                        self.state = 57
                        self.match(RAParser.OR)
                        self.state = 58
                        self.valExpr(2)
                        pass

                    elif la_ == 8:
                        localctx = RAParser.IsNullExprContext(self, RAParser.ValExprContext(self, _parentctx, _parentState))
                        self.pushNewRecursionContext(localctx, _startState, self.RULE_valExpr)
                        self.state = 59
                        if not self.precpred(self._ctx, 5):
                            from antlr4.error.Errors import FailedPredicateException
                            raise FailedPredicateException(self, "self.precpred(self._ctx, 5)")
                        self.state = 60
                        self.match(RAParser.IS_NULL)
                        pass

                    elif la_ == 9:
                        localctx = RAParser.IsNotNullExprContext(self, RAParser.ValExprContext(self, _parentctx, _parentState))
                        self.pushNewRecursionContext(localctx, _startState, self.RULE_valExpr)
                        self.state = 61
                        if not self.precpred(self._ctx, 4):
                            from antlr4.error.Errors import FailedPredicateException
                            raise FailedPredicateException(self, "self.precpred(self._ctx, 4)")
                        self.state = 62
                        self.match(RAParser.IS_NOT_NULL)
                        pass

             
                self.state = 67
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input,4,self._ctx)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.unrollRecursionContexts(_parentctx)
        return localctx

    class ListOfValExprsContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def valExpr(self):
            return self.getTypedRuleContext(RAParser.ValExprContext,0)


        def COMMA(self):
            return self.getToken(RAParser.COMMA, 0)

        def listOfValExprs(self):
            return self.getTypedRuleContext(RAParser.ListOfValExprsContext,0)


        def getRuleIndex(self):
            return RAParser.RULE_listOfValExprs

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitListOfValExprs" ):
                return visitor.visitListOfValExprs(self)
            else:
                return visitor.visitChildren(self)




    def listOfValExprs(self):

        localctx = RAParser.ListOfValExprsContext(self, self._ctx, self.state)
        self.enterRule(localctx, 2, self.RULE_listOfValExprs)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 68
            self.valExpr(0)
            self.state = 71
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==RAParser.COMMA:
                self.state = 69
                self.match(RAParser.COMMA)
                self.state = 70
                self.listOfValExprs()


        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class ListOfIDsContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def ID(self):
            return self.getToken(RAParser.ID, 0)

        def COMMA(self):
            return self.getToken(RAParser.COMMA, 0)

        def listOfIDs(self):
            return self.getTypedRuleContext(RAParser.ListOfIDsContext,0)


        def getRuleIndex(self):
            return RAParser.RULE_listOfIDs

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitListOfIDs" ):
                return visitor.visitListOfIDs(self)
            else:
                return visitor.visitChildren(self)




    def listOfIDs(self):

        localctx = RAParser.ListOfIDsContext(self, self._ctx, self.state)
        self.enterRule(localctx, 4, self.RULE_listOfIDs)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 73
            self.match(RAParser.ID)
            self.state = 76
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==RAParser.COMMA:
                self.state = 74
                self.match(RAParser.COMMA)
                self.state = 75
                self.listOfIDs()


        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class RelExprContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser


        def getRuleIndex(self):
            return RAParser.RULE_relExpr

     
        def copyFrom(self, ctx:ParserRuleContext):
            super().copyFrom(ctx)


    class RelExprParenthesizedContext(RelExprContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a RAParser.RelExprContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def PAREN_L(self):
            return self.getToken(RAParser.PAREN_L, 0)
        def relExpr(self):
            return self.getTypedRuleContext(RAParser.RelExprContext,0)

        def PAREN_R(self):
            return self.getToken(RAParser.PAREN_R, 0)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitRelExprParenthesized" ):
                return visitor.visitRelExprParenthesized(self)
            else:
                return visitor.visitChildren(self)


    class CrossExprContext(RelExprContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a RAParser.RelExprContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def relExpr(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(RAParser.RelExprContext)
            else:
                return self.getTypedRuleContext(RAParser.RelExprContext,i)

        def CROSS(self):
            return self.getToken(RAParser.CROSS, 0)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitCrossExpr" ):
                return visitor.visitCrossExpr(self)
            else:
                return visitor.visitChildren(self)


    class ProjectExprContext(RelExprContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a RAParser.RelExprContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def PROJECT(self):
            return self.getToken(RAParser.PROJECT, 0)
        def ARG_L(self):
            return self.getToken(RAParser.ARG_L, 0)
        def listOfValExprs(self):
            return self.getTypedRuleContext(RAParser.ListOfValExprsContext,0)

        def ARG_R(self):
            return self.getToken(RAParser.ARG_R, 0)
        def relExpr(self):
            return self.getTypedRuleContext(RAParser.RelExprContext,0)


        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitProjectExpr" ):
                return visitor.visitProjectExpr(self)
            else:
                return visitor.visitChildren(self)


    class RelRefContext(RelExprContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a RAParser.RelExprContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def ID(self):
            return self.getToken(RAParser.ID, 0)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitRelRef" ):
                return visitor.visitRelRef(self)
            else:
                return visitor.visitChildren(self)


    class SelectExprContext(RelExprContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a RAParser.RelExprContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def SELECT(self):
            return self.getToken(RAParser.SELECT, 0)
        def ARG_L(self):
            return self.getToken(RAParser.ARG_L, 0)
        def valExpr(self):
            return self.getTypedRuleContext(RAParser.ValExprContext,0)

        def ARG_R(self):
            return self.getToken(RAParser.ARG_R, 0)
        def relExpr(self):
            return self.getTypedRuleContext(RAParser.RelExprContext,0)


        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitSelectExpr" ):
                return visitor.visitSelectExpr(self)
            else:
                return visitor.visitChildren(self)


    class IntersectExprContext(RelExprContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a RAParser.RelExprContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def relExpr(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(RAParser.RelExprContext)
            else:
                return self.getTypedRuleContext(RAParser.RelExprContext,i)

        def INTERSECT(self):
            return self.getToken(RAParser.INTERSECT, 0)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitIntersectExpr" ):
                return visitor.visitIntersectExpr(self)
            else:
                return visitor.visitChildren(self)


    class DiffExprContext(RelExprContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a RAParser.RelExprContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def relExpr(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(RAParser.RelExprContext)
            else:
                return self.getTypedRuleContext(RAParser.RelExprContext,i)

        def DIFF(self):
            return self.getToken(RAParser.DIFF, 0)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitDiffExpr" ):
                return visitor.visitDiffExpr(self)
            else:
                return visitor.visitChildren(self)


    class RenameExprContext(RelExprContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a RAParser.RelExprContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def RENAME(self):
            return self.getToken(RAParser.RENAME, 0)
        def ARG_L(self):
            return self.getToken(RAParser.ARG_L, 0)
        def ARG_R(self):
            return self.getToken(RAParser.ARG_R, 0)
        def relExpr(self):
            return self.getTypedRuleContext(RAParser.RelExprContext,0)

        def listOfIDs(self):
            return self.getTypedRuleContext(RAParser.ListOfIDsContext,0)

        def ID(self):
            return self.getToken(RAParser.ID, 0)
        def COLON(self):
            return self.getToken(RAParser.COLON, 0)
        def STAR(self):
            return self.getToken(RAParser.STAR, 0)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitRenameExpr" ):
                return visitor.visitRenameExpr(self)
            else:
                return visitor.visitChildren(self)


    class JoinExprContext(RelExprContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a RAParser.RelExprContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def relExpr(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(RAParser.RelExprContext)
            else:
                return self.getTypedRuleContext(RAParser.RelExprContext,i)

        def JOIN(self):
            return self.getToken(RAParser.JOIN, 0)
        def ARG_L(self):
            return self.getToken(RAParser.ARG_L, 0)
        def valExpr(self):
            return self.getTypedRuleContext(RAParser.ValExprContext,0)

        def ARG_R(self):
            return self.getToken(RAParser.ARG_R, 0)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitJoinExpr" ):
                return visitor.visitJoinExpr(self)
            else:
                return visitor.visitChildren(self)


    class UnionExprContext(RelExprContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a RAParser.RelExprContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def relExpr(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(RAParser.RelExprContext)
            else:
                return self.getTypedRuleContext(RAParser.RelExprContext,i)

        def UNION(self):
            return self.getToken(RAParser.UNION, 0)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitUnionExpr" ):
                return visitor.visitUnionExpr(self)
            else:
                return visitor.visitChildren(self)



    def relExpr(self, _p:int=0):
        _parentctx = self._ctx
        _parentState = self.state
        localctx = RAParser.RelExprContext(self, self._ctx, _parentState)
        _prevctx = localctx
        _startState = 6
        self.enterRecursionRule(localctx, 6, self.RULE_relExpr, _p)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 109
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [RAParser.PAREN_L]:
                localctx = RAParser.RelExprParenthesizedContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx

                self.state = 79
                self.match(RAParser.PAREN_L)
                self.state = 80
                self.relExpr(0)
                self.state = 81
                self.match(RAParser.PAREN_R)
                pass
            elif token in [RAParser.ID]:
                localctx = RAParser.RelRefContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx
                self.state = 83
                self.match(RAParser.ID)
                pass
            elif token in [RAParser.RENAME]:
                localctx = RAParser.RenameExprContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx
                self.state = 84
                self.match(RAParser.RENAME)
                self.state = 85
                self.match(RAParser.ARG_L)
                self.state = 93
                self._errHandler.sync(self)
                la_ = self._interp.adaptivePredict(self._input,8,self._ctx)
                if la_ == 1:
                    self.state = 86
                    self.match(RAParser.ID)
                    self.state = 87
                    self.match(RAParser.COLON)
                    self.state = 90
                    self._errHandler.sync(self)
                    token = self._input.LA(1)
                    if token in [RAParser.STAR]:
                        self.state = 88
                        self.match(RAParser.STAR)
                        pass
                    elif token in [RAParser.ID]:
                        self.state = 89
                        self.listOfIDs()
                        pass
                    else:
                        raise NoViableAltException(self)

                    pass

                elif la_ == 2:
                    self.state = 92
                    self.listOfIDs()
                    pass


                self.state = 95
                self.match(RAParser.ARG_R)
                self.state = 96
                self.relExpr(8)
                pass
            elif token in [RAParser.PROJECT]:
                localctx = RAParser.ProjectExprContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx
                self.state = 97
                self.match(RAParser.PROJECT)
                self.state = 98
                self.match(RAParser.ARG_L)
                self.state = 99
                self.listOfValExprs()
                self.state = 100
                self.match(RAParser.ARG_R)
                self.state = 101
                self.relExpr(7)
                pass
            elif token in [RAParser.SELECT]:
                localctx = RAParser.SelectExprContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx
                self.state = 103
                self.match(RAParser.SELECT)
                self.state = 104
                self.match(RAParser.ARG_L)
                self.state = 105
                self.valExpr(0)
                self.state = 106
                self.match(RAParser.ARG_R)
                self.state = 107
                self.relExpr(6)
                pass
            else:
                raise NoViableAltException(self)

            self._ctx.stop = self._input.LT(-1)
            self.state = 134
            self._errHandler.sync(self)
            _alt = self._interp.adaptivePredict(self._input,12,self._ctx)
            while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:
                if _alt==1:
                    if self._parseListeners is not None:
                        self.triggerExitRuleEvent()
                    _prevctx = localctx
                    self.state = 132
                    self._errHandler.sync(self)
                    la_ = self._interp.adaptivePredict(self._input,11,self._ctx)
                    if la_ == 1:
                        localctx = RAParser.JoinExprContext(self, RAParser.RelExprContext(self, _parentctx, _parentState))
                        self.pushNewRecursionContext(localctx, _startState, self.RULE_relExpr)
                        self.state = 111
                        if not self.precpred(self._ctx, 5):
                            from antlr4.error.Errors import FailedPredicateException
                            raise FailedPredicateException(self, "self.precpred(self._ctx, 5)")
                        self.state = 112
                        self.match(RAParser.JOIN)
                        self.state = 117
                        self._errHandler.sync(self)
                        _la = self._input.LA(1)
                        if _la==RAParser.ARG_L:
                            self.state = 113
                            self.match(RAParser.ARG_L)
                            self.state = 114
                            self.valExpr(0)
                            self.state = 115
                            self.match(RAParser.ARG_R)


                        self.state = 119
                        self.relExpr(6)
                        pass

                    elif la_ == 2:
                        localctx = RAParser.CrossExprContext(self, RAParser.RelExprContext(self, _parentctx, _parentState))
                        self.pushNewRecursionContext(localctx, _startState, self.RULE_relExpr)
                        self.state = 120
                        if not self.precpred(self._ctx, 4):
                            from antlr4.error.Errors import FailedPredicateException
                            raise FailedPredicateException(self, "self.precpred(self._ctx, 4)")
                        self.state = 121
                        self.match(RAParser.CROSS)
                        self.state = 122
                        self.relExpr(5)
                        pass

                    elif la_ == 3:
                        localctx = RAParser.UnionExprContext(self, RAParser.RelExprContext(self, _parentctx, _parentState))
                        self.pushNewRecursionContext(localctx, _startState, self.RULE_relExpr)
                        self.state = 123
                        if not self.precpred(self._ctx, 3):
                            from antlr4.error.Errors import FailedPredicateException
                            raise FailedPredicateException(self, "self.precpred(self._ctx, 3)")
                        self.state = 124
                        self.match(RAParser.UNION)
                        self.state = 125
                        self.relExpr(4)
                        pass

                    elif la_ == 4:
                        localctx = RAParser.DiffExprContext(self, RAParser.RelExprContext(self, _parentctx, _parentState))
                        self.pushNewRecursionContext(localctx, _startState, self.RULE_relExpr)
                        self.state = 126
                        if not self.precpred(self._ctx, 2):
                            from antlr4.error.Errors import FailedPredicateException
                            raise FailedPredicateException(self, "self.precpred(self._ctx, 2)")
                        self.state = 127
                        self.match(RAParser.DIFF)
                        self.state = 128
                        self.relExpr(3)
                        pass

                    elif la_ == 5:
                        localctx = RAParser.IntersectExprContext(self, RAParser.RelExprContext(self, _parentctx, _parentState))
                        self.pushNewRecursionContext(localctx, _startState, self.RULE_relExpr)
                        self.state = 129
                        if not self.precpred(self._ctx, 1):
                            from antlr4.error.Errors import FailedPredicateException
                            raise FailedPredicateException(self, "self.precpred(self._ctx, 1)")
                        self.state = 130
                        self.match(RAParser.INTERSECT)
                        self.state = 131
                        self.relExpr(2)
                        pass

             
                self.state = 136
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input,12,self._ctx)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.unrollRecursionContexts(_parentctx)
        return localctx

    class DefinitionContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def ID(self):
            return self.getToken(RAParser.ID, 0)

        def GETS(self):
            return self.getToken(RAParser.GETS, 0)

        def relExpr(self):
            return self.getTypedRuleContext(RAParser.RelExprContext,0)


        def getRuleIndex(self):
            return RAParser.RULE_definition

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitDefinition" ):
                return visitor.visitDefinition(self)
            else:
                return visitor.visitChildren(self)




    def definition(self):

        localctx = RAParser.DefinitionContext(self, self._ctx, self.state)
        self.enterRule(localctx, 8, self.RULE_definition)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 137
            self.match(RAParser.ID)
            self.state = 138
            self.match(RAParser.GETS)
            self.state = 139
            self.relExpr(0)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class CommandContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser


        def getRuleIndex(self):
            return RAParser.RULE_command

     
        def copyFrom(self, ctx:ParserRuleContext):
            super().copyFrom(ctx)



    class QuitCommandContext(CommandContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a RAParser.CommandContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def QUIT(self):
            return self.getToken(RAParser.QUIT, 0)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitQuitCommand" ):
                return visitor.visitQuitCommand(self)
            else:
                return visitor.visitChildren(self)


    class ClearCommandContext(CommandContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a RAParser.CommandContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def CLEAR(self):
            return self.getToken(RAParser.CLEAR, 0)
        def ID(self):
            return self.getToken(RAParser.ID, 0)
        def STAR(self):
            return self.getToken(RAParser.STAR, 0)
        def FORCE(self):
            return self.getToken(RAParser.FORCE, 0)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitClearCommand" ):
                return visitor.visitClearCommand(self)
            else:
                return visitor.visitChildren(self)


    class SqlexecCommandContext(CommandContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a RAParser.CommandContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def SQLEXEC(self):
            return self.getToken(RAParser.SQLEXEC, 0)
        def SQLEXEC_TEXT(self):
            return self.getToken(RAParser.SQLEXEC_TEXT, 0)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitSqlexecCommand" ):
                return visitor.visitSqlexecCommand(self)
            else:
                return visitor.visitChildren(self)


    class SaveCommandContext(CommandContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a RAParser.CommandContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def SAVE(self):
            return self.getToken(RAParser.SAVE, 0)
        def ID(self):
            return self.getToken(RAParser.ID, 0)
        def STAR(self):
            return self.getToken(RAParser.STAR, 0)
        def FORCE(self):
            return self.getToken(RAParser.FORCE, 0)
        def STRING(self):
            return self.getToken(RAParser.STRING, 0)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitSaveCommand" ):
                return visitor.visitSaveCommand(self)
            else:
                return visitor.visitChildren(self)


    class SourceCommandContext(CommandContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a RAParser.CommandContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def SOURCE(self):
            return self.getToken(RAParser.SOURCE, 0)
        def STRING(self):
            return self.getToken(RAParser.STRING, 0)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitSourceCommand" ):
                return visitor.visitSourceCommand(self)
            else:
                return visitor.visitChildren(self)


    class ListCommandContext(CommandContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a RAParser.CommandContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def LIST(self):
            return self.getToken(RAParser.LIST, 0)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitListCommand" ):
                return visitor.visitListCommand(self)
            else:
                return visitor.visitChildren(self)



    def command(self):

        localctx = RAParser.CommandContext(self, self._ctx, self.state)
        self.enterRule(localctx, 10, self.RULE_command)
        self._la = 0 # Token type
        try:
            self.state = 163
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [RAParser.LIST]:
                localctx = RAParser.ListCommandContext(self, localctx)
                self.enterOuterAlt(localctx, 1)
                self.state = 141
                self.match(RAParser.LIST)
                pass
            elif token in [RAParser.CLEAR]:
                localctx = RAParser.ClearCommandContext(self, localctx)
                self.enterOuterAlt(localctx, 2)
                self.state = 142
                self.match(RAParser.CLEAR)
                self.state = 148
                self._errHandler.sync(self)
                token = self._input.LA(1)
                if token in [RAParser.FORCE, RAParser.ID]:
                    self.state = 144
                    self._errHandler.sync(self)
                    _la = self._input.LA(1)
                    if _la==RAParser.FORCE:
                        self.state = 143
                        self.match(RAParser.FORCE)


                    self.state = 146
                    self.match(RAParser.ID)
                    pass
                elif token in [RAParser.STAR]:
                    self.state = 147
                    self.match(RAParser.STAR)
                    pass
                else:
                    raise NoViableAltException(self)

                pass
            elif token in [RAParser.SAVE]:
                localctx = RAParser.SaveCommandContext(self, localctx)
                self.enterOuterAlt(localctx, 3)
                self.state = 150
                self.match(RAParser.SAVE)
                self.state = 152
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la==RAParser.FORCE:
                    self.state = 151
                    self.match(RAParser.FORCE)


                self.state = 154
                _la = self._input.LA(1)
                if not(_la==RAParser.ID or _la==RAParser.STAR):
                    self._errHandler.recoverInline(self)
                else:
                    self._errHandler.reportMatch(self)
                    self.consume()
                self.state = 156
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la==RAParser.STRING:
                    self.state = 155
                    self.match(RAParser.STRING)


                pass
            elif token in [RAParser.SOURCE]:
                localctx = RAParser.SourceCommandContext(self, localctx)
                self.enterOuterAlt(localctx, 4)
                self.state = 158
                self.match(RAParser.SOURCE)
                self.state = 159
                self.match(RAParser.STRING)
                pass
            elif token in [RAParser.QUIT]:
                localctx = RAParser.QuitCommandContext(self, localctx)
                self.enterOuterAlt(localctx, 5)
                self.state = 160
                self.match(RAParser.QUIT)
                pass
            elif token in [RAParser.SQLEXEC]:
                localctx = RAParser.SqlexecCommandContext(self, localctx)
                self.enterOuterAlt(localctx, 6)
                self.state = 161
                self.match(RAParser.SQLEXEC)
                self.state = 162
                self.match(RAParser.SQLEXEC_TEXT)
                pass
            else:
                raise NoViableAltException(self)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class StatementContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def TERMINATOR(self):
            return self.getToken(RAParser.TERMINATOR, 0)

        def relExpr(self):
            return self.getTypedRuleContext(RAParser.RelExprContext,0)


        def definition(self):
            return self.getTypedRuleContext(RAParser.DefinitionContext,0)


        def command(self):
            return self.getTypedRuleContext(RAParser.CommandContext,0)


        def getRuleIndex(self):
            return RAParser.RULE_statement

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitStatement" ):
                return visitor.visitStatement(self)
            else:
                return visitor.visitChildren(self)




    def statement(self):

        localctx = RAParser.StatementContext(self, self._ctx, self.state)
        self.enterRule(localctx, 12, self.RULE_statement)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 168
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input,18,self._ctx)
            if la_ == 1:
                self.state = 165
                self.relExpr(0)
                pass

            elif la_ == 2:
                self.state = 166
                self.definition()
                pass

            elif la_ == 3:
                self.state = 167
                self.command()
                pass


            self.state = 170
            self.match(RAParser.TERMINATOR)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class ProgramContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def EOF(self):
            return self.getToken(RAParser.EOF, 0)

        def statement(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(RAParser.StatementContext)
            else:
                return self.getTypedRuleContext(RAParser.StatementContext,i)


        def getRuleIndex(self):
            return RAParser.RULE_program

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitProgram" ):
                return visitor.visitProgram(self)
            else:
                return visitor.visitChildren(self)




    def program(self):

        localctx = RAParser.ProgramContext(self, self._ctx, self.state)
        self.enterRule(localctx, 14, self.RULE_program)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 175
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << RAParser.ID) | (1 << RAParser.RENAME) | (1 << RAParser.PROJECT) | (1 << RAParser.SELECT) | (1 << RAParser.PAREN_L) | (1 << RAParser.LIST) | (1 << RAParser.CLEAR) | (1 << RAParser.SAVE) | (1 << RAParser.SOURCE) | (1 << RAParser.QUIT) | (1 << RAParser.SQLEXEC))) != 0):
                self.state = 172
                self.statement()
                self.state = 177
                self._errHandler.sync(self)
                _la = self._input.LA(1)

            self.state = 178
            self.match(RAParser.EOF)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx



    def sempred(self, localctx:RuleContext, ruleIndex:int, predIndex:int):
        if self._predicates == None:
            self._predicates = dict()
        self._predicates[0] = self.valExpr_sempred
        self._predicates[3] = self.relExpr_sempred
        pred = self._predicates.get(ruleIndex, None)
        if pred is None:
            raise Exception("No predicate with index:" + str(ruleIndex))
        else:
            return pred(localctx, predIndex)

    def valExpr_sempred(self, localctx:ValExprContext, predIndex:int):
            if predIndex == 0:
                return self.precpred(self._ctx, 10)
         

            if predIndex == 1:
                return self.precpred(self._ctx, 9)
         

            if predIndex == 2:
                return self.precpred(self._ctx, 8)
         

            if predIndex == 3:
                return self.precpred(self._ctx, 7)
         

            if predIndex == 4:
                return self.precpred(self._ctx, 6)
         

            if predIndex == 5:
                return self.precpred(self._ctx, 2)
         

            if predIndex == 6:
                return self.precpred(self._ctx, 1)
         

            if predIndex == 7:
                return self.precpred(self._ctx, 5)
         

            if predIndex == 8:
                return self.precpred(self._ctx, 4)
         

    def relExpr_sempred(self, localctx:RelExprContext, predIndex:int):
            if predIndex == 9:
                return self.precpred(self._ctx, 5)
         

            if predIndex == 10:
                return self.precpred(self._ctx, 4)
         

            if predIndex == 11:
                return self.precpred(self._ctx, 3)
         

            if predIndex == 12:
                return self.precpred(self._ctx, 2)
         

            if predIndex == 13:
                return self.precpred(self._ctx, 1)
         







import os
import sys
import configparser
import argparse
import atexit
import getpass

try:
    import readline
except ImportError:
    _readline_available = False
else:
    _readline_available = True

from radb.db import DB
from radb.parse import ParsingError,\
    statement_string_from_stdin, one_statement_from_string, RACompleter
from radb.typesys import ValTypeChecker, TypeSysError
from radb.views import ViewCollection
from radb import utils
from radb.ast import Context, ValidationError, ExecutionError, execute_from_file

import logging
logger = logging.getLogger('ra')

def main():

    # read system defaults:
    sys_configfile = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'sys.ini')
    sys_config = configparser.ConfigParser()
    if sys_config.read(sys_configfile) == []:
        print('ERROR: required system configuration file {} not found'.format(sys_configfile))
        sys.exit(1)
    defaults = dict(sys_config.items(configparser.DEFAULTSECT))

    # parse input arguments:
    parser = argparse.ArgumentParser()
    parser.add_argument('--configfile', '-c', type=str, default=defaults['configfile'],
                        help='configuration file (default: {})'.format(defaults['configfile']))
    parser.add_argument('--password', '-p', action='store_true',
                        help='prompt for database password')
    parser.add_argument('--inputfile', '-i', type=str,
                        help='input file')
    parser.add_argument('--outputfile', '-o', type=str,
                        help='output file')
    parser.add_argument('--echo', '-e', action='store_true',
                        help='echo input')
    parser.add_argument('--verbose', '-v', action='store_true',
                        help='verbose output')
    parser.add_argument('--debug', '-d', action='store_true',
                        help='debug output')
    parser.add_argument('source', type=str, nargs='?', default=configparser.DEFAULTSECT,
                        help='data source, which can be the name of a configuration section'
                        ' (default: {}), or otherwise the name of the database to connect to'
                        ' (overriding the configuration default)'.format(configparser.DEFAULTSECT))
    args = parser.parse_args()

    # set up output replication + logging:
    if args.outputfile is not None:
        sys.stdout = utils.Tee(args.outputfile)
    logging.getLogger('ra').setLevel(logging.DEBUG if args.debug else\
                                     (logging.INFO if args.verbose else\
                                      logging.WARNING))
    logger_handler = logging.StreamHandler(sys.stdout)
    logger_handler.setFormatter(logging.Formatter('%(levelname)s: %(message)s'))
    logger.addHandler(logger_handler)

    # read user configuration file (starting with system defaults):
    config = configparser.ConfigParser(defaults)
    if config.read(os.path.expanduser(args.configfile)) == []:
        logger.warning('unable to read configuration file {}; resorting to system defaults'\
                       .format(os.path.expanduser(args.configfile)))

    # finalize configuration settings, using configuration file and command-line arguments:
    if args.source == configparser.DEFAULTSECT or config.has_section(args.source):
        configured = dict(config.items(args.source))
    else: # args.source is not a section in the config file; treat it as a database name:
        configured = dict(config.items(configparser.DEFAULTSECT))
        configured['db.database'] = args.source
    if args.password:
        configured['db.password'] = getpass.getpass('Database password: ')

    # connect to database:
    if 'db.database' not in configured:
        logger.warning('no database specified')
    try:
        db = DB(configured)
    except Exception as e:
        logger.error('failed to connect to database: {}'.format(e))
        sys.exit(1)

    # initialize type system:
    try:
        check = ValTypeChecker(configured['default_functions'], configured.get('functions', None))
    except TypeSysError as e:
        logger.error(e)
        sys.exit(1)

    # construct context (starting with empty view collection):
    context = Context(db, check, ViewCollection())

    # finally, let's start:
    if args.inputfile is None: # interactive:
        if _readline_available:
            # set up command history file:
            historyfile = os.path.expanduser(configured['historyfile'])
            try:
                readline.read_history_file(historyfile)
                history_length = readline.get_history_length()
            except FileNotFoundError:
                open(historyfile, 'wb').close()
                history_length = 0
            def save_history(prev_history_length, historyfile):
                new_history_length = readline.get_current_history_length()
                readline.set_history_length(1000)
                readline.append_history_file(new_history_length - prev_history_length, historyfile)
            atexit.register(save_history, history_length, historyfile)
            # set up command completion:
            readline.set_completer_delims(readline.get_completer_delims().replace('\\', ''))
            readline.parse_and_bind('tab: complete')
            readline.set_completer(RACompleter().complete)
            atexit.register(lambda: readline.set_completer(None))
        print('{name}: {description}\nVersion {version} by {author} <{author_email}>\n{url}'.
              format(**{key: configured['setup.' + key]\
                        for key in ('name', 'description',
                                    'version', 'author', 'author_email',
                                    'url')}))
        for s in statement_string_from_stdin(echo=args.echo):
            logger.info('statement received:')
            for line in utils.number_lines(s):
                logger.info(line)
            try:
                ast = one_statement_from_string(s)
                logger.info('statement parsed:')
                logger.info(str(ast))
                ast.validate(context)
                logger.info('statement validated:')
                for line in ast.info():
                    logger.info(line)
                ast.execute(context)
            except (ParsingError, ValidationError, ExecutionError) as e:
                logger.error(e)
    else:
        try:
            execute_from_file(args.inputfile, context, echo=args.echo)
        except (IOError, ParsingError, ValidationError, ExecutionError) as e:
            logger.error(e)
            sys.exit(1)

if __name__ == '__main__':
    main()



import os
import sys
import itertools

from radb.utils import CustomJSONSerializable, lreplace
from radb.typesys import ValType, AttrSpec, RelType, TypeSysError
from radb.parse import RAParser as sym
from radb.parse import literal, symbolic, sqlstr_to_str, str_to_sqlstr
from radb.parse import statements_from_file, one_statement_from_string

import logging
logger = logging.getLogger('ra')

######################################################################

class Context:
    def __init__(self, db, check, views):
        self.db = db
        self.check = check
        self.views = views

class StatementContext(Context):
    def __init__(self, context: Context, root: 'RelType'):
        super(StatementContext, self).__init__(context.db, context.check, context.views)
        self.tmps = 0
        self.root = root
    def new_tmp(self):
        self.tmps += 1
        return self.tmps-1

######################################################################

def execute_from_file(filename, context, echo=False):
    for ast in statements_from_file(filename):
        if echo:
            print(str(ast) + literal(sym.TERMINATOR))
        ast.validate(context)
        logger.info('statement validated:')
        for line in ast.info():
            logger.info(line)
        ast.execute(context)

class ExecutionError(Exception):
    pass

class ValidationError(Exception):
    def __init__(self, message, node, relexpr=None):
        assert isinstance(message, str)
        self.message = message
        assert isinstance(node, Node)
        self.node = node
        assert relexpr is None or isinstance(relexpr, RelExpr)
        self.relexpr = relexpr
        if self.relexpr is not None and self.relexpr != self.node:
            full_message = 'in {}:\n{}\ncontext: {}'.format(self.node, self.message, self.relexpr)
        else:
            full_message = 'in {}:\n{}'.format(self.node, self.message)
        super(Exception, self).__init__(full_message)

######################################################################

class Node(CustomJSONSerializable):
    pass

class paren: # a decorator for ValExpr and RelExpr
    def __init__(self, expr):
        assert isinstance(expr, ValExpr) or isinstance(expr, RelExpr)
        self.expr = expr
    def can_omit_paren(self):
        return any(isinstance(self.expr, cls) for cls in (Literal, AttrRef, RelRef))
    def __str__(self):
        return ('{}' if self.can_omit_paren() else\
                literal(sym.PAREN_L) + '{}' + literal(sym.PAREN_R))\
                .format(self.expr)
    def info(self):
        return ('{}' if self.can_omit_paren() else '({})').format(self.expr.info())
    def sql(self, relexpr=None):
        return ('{}' if self.can_omit_paren() else '({})')\
            .format(self.expr.sql() if relexpr is None else self.expr.sql(relexpr))

######################################################################

class ValExpr(Node):
    def __init__(self, inputs=list()):
        assert isinstance(inputs, list)
        assert all(isinstance(input, ValExpr) for input in inputs)
        self.inputs = inputs
    def validateSubtree(self, context: StatementContext, relexpr: 'RelExpr'):
        """Validate this value expression (in the scope of the given RelExpr)
        and set its type (as a ValType), using the input RelTypes of
        the RelExpr to resolve attribute references.
        """
        raise NotImplementedError
    def info(self):
        raise NotImplementedError
    def sql(self, relexpr):
        raise NotImplementedError

class Literal(ValExpr):
    def info(self):
        return str(self)
    def sql(self, relexpr):
        return str(self)

class RAString(Literal):
    def __init__(self, val):
        super(RAString, self).__init__()
        assert isinstance(val, str) # include SQL quotes and escapes
        self.val = val
    def __str__(self):
        return self.val
    def validateSubtree(self, context: StatementContext, relexpr):
        self.type = ValType.STRING

class RANumber(Literal):
    def __init__(self, val):
        super(RANumber, self).__init__()
        assert isinstance(val, str) # still represent as a string
        self.val = val
    def __str__(self):
        return self.val
    def validateSubtree(self, context: StatementContext, relexpr):
        self.type = ValType.NUMBER

class FuncValExpr(ValExpr):
    def __init__(self, func, args):
        super(FuncValExpr, self).__init__()
        assert isinstance(func, str)
        self.func = func
        assert isinstance(args, list) and all(isinstance(arg, ValExpr) for arg in args)
        self.args = args
    def __str__(self):
        return self.func + literal(sym.PAREN_L) +\
            ', '.join(str(arg) for arg in self.args) +\
            literal(sym.PAREN_R)
    def validateSubtree(self, context: StatementContext, relexpr):
        argtypes = list()
        for arg in self.args:
            arg.validateSubtree(context, relexpr)
            argtypes.append(arg.type)
        try:
            self.type = context.check.function_call(self.func, argtypes)
        except TypeSysError as e:
            raise ValidationError(str(e), self, relexpr) from e
    def info(self):
        return self.func + literal(sym.PAREN_L) +\
            ', '.join(arg.info() for arg in self.args) +\
            literal(sym.PAREN_R)
    def sql(self, relexpr):
        return self.func + literal(sym.PAREN_L) +\
            ', '.join(arg.sql(relexpr) for arg in self.args) +\
            literal(sym.PAREN_R)

class AttrRef(ValExpr):
    def __init__(self, rel, name):
        super(AttrRef, self).__init__()
        assert rel is None or isinstance(rel, str)
        assert isinstance(name, str)
        self.rel = rel
        self.name = name
    def __str__(self):
        if self.rel is None:
            return self.name
        else:
            return self.rel + literal(sym.DOT) + self.name
    def validateSubtree(self, context: StatementContext, relexpr):
        """Validate this attribute reference (in the scope of the given
        RelExpr) and set its type (as a ValType) and internal
        reference (used in translation), using the input RelTypes of
        the RelExpr to resolve attribute references.
        """
        self.type = None
        self.internal_ref = None
        for ridx, relinput in enumerate(relexpr.inputs):
            for aidx, attrspec in enumerate(relinput.type.attrs):
                if self.name == attrspec.name and\
                   (self.rel is None or self.rel == attrspec.rel):
                    if self.internal_ref is not None:
                        raise ValidationError('ambiguous attribute reference', self, relexpr)
                    self.type = attrspec.type
                    self.internal_ref = (ridx, aidx)
        if self.internal_ref is None:
            raise ValidationError('invalid attribute reference', self, relexpr)
    def info(self):
        return '{}[{}.{}]'.format(self, *self.internal_ref)
    def sql(self, relexpr):
        ridx, aidx = self.internal_ref
        return '{}.{}'.format(relexpr.inputs[ridx].type.sql_rel(),
                              relexpr.inputs[ridx].type.sql_attr(aidx))

class ValExprBinaryOp(ValExpr):
    def __init__(self, left, op, right):
        assert isinstance(left, ValExpr)
        assert op in (sym.STAR, sym.SLASH, sym.PLUS, sym.MINUS,
                      sym.LT, sym.LE, sym.EQ, sym.NE, sym.GE, sym.GT,
                      sym.CONCAT,
                      sym.AND, sym.OR,
                      sym.LIKE)
        assert isinstance(right, ValExpr)
        super(ValExprBinaryOp, self).__init__([left, right])
        self.op = op
    def __str__(self):
        return '{} {} {}'.format(paren(self.inputs[0]), literal(self.op), paren(self.inputs[1]))
    def validateSubtree(self, context: StatementContext, relexpr):
        self.inputs[0].validateSubtree(context, relexpr)
        self.inputs[1].validateSubtree(context, relexpr)
        try:
            self.type = context.check.function_call(symbolic(self.op),
                                                    [self.inputs[0].type, self.inputs[1].type])
        except TypeSysError as e:
            raise ValidationError(str(e), self, relexpr) from e
    def info(self):
        return '{} {} {}'.format(paren(self.inputs[0]).info(),
                                 literal(self.op),
                                 paren(self.inputs[1]).info())
    def sql(self, relexpr):
        return '{} {} {}'.format(paren(self.inputs[0]).sql(relexpr),
                                 literal(self.op),
                                 paren(self.inputs[1]).sql(relexpr))

class ValExprUnaryOp(ValExpr):
    def __init__(self, op, input):
        assert isinstance(input, ValExpr)
        assert op in (sym.NOT, sym.IS_NULL, sym.IS_NOT_NULL)
        self.op = op
        super(ValExprUnaryOp, self).__init__([input])
    def __str__(self):
        fmt = '{input} {op}' if self.op in (sym.IS_NULL, sym.IS_NOT_NULL) else '{op} {input}'
        return fmt.format(op=literal(self.op), input=paren(self.inputs[0]))
    def validateSubtree(self, context: StatementContext, relexpr):
        self.inputs[0].validateSubtree(context, relexpr)
        try:
            self.type = context.check.function_call(symbolic(self.op), [self.inputs[0].type])
        except TypeSysError as e:
            raise ValidationError(str(e), self, relexpr) from e
    def info(self):
        fmt = '{input} {op}' if self.op in (sym.IS_NULL, sym.IS_NOT_NULL) else '{op} {input}'
        return fmt.format(op=literal(self.op), input=paren(self.inputs[0]).info())
    def sql(self, relexpr):
        fmt = '{input} {op}' if self.op in (sym.IS_NULL, sym.IS_NOT_NULL) else '{op} {input}'
        return fmt.format(op=literal(self.op), input=paren(self.inputs[0]).sql(relexpr))

######################################################################

class RelExpr(Node):
    def __init__(self, inputs=list()):
        assert isinstance(inputs, list)
        assert all(isinstance(input, RelExpr) for input in inputs)
        self.inputs = inputs
    def validate(self, context: Context):
        self.validateSubtree(StatementContext(context, self))
    def validateSubtree(self, context: StatementContext):
        """Validate this value expression and set its type (as a RelType),
        using the given context to resolve relation references.
        """
        pass
    def validationError(self, message):
        return ValidationError(message, self, self)
    def find_views(self):
        if isinstance(self, RelRef):
            return set() if self.view is None else {self.rel}
        else:
            views = set()
            for input in self.inputs:
                views |= input.find_views()
            return views
    def info(self):
        raise NotImplementedError
    def sql(self):
        raise NotImplementedError
    def execute(self, context):
        blocks = [block for block in self.sql()]
        assert(len(blocks) > 0)
        query = 'WITH ' + ',\n     '.join(blocks)
        query += '\nSELECT * FROM {}'.format(self.type.sql_rel())
        logger.debug('SQL generated:\n' + query)
        try:
            print('({})'.format(', '.join(self.type.str_attr_names_and_types())))
            context.db.execute_and_print_result(query)
        except Exception as e:
            raise ExecutionError('SQL error in translated query:\n{}\n{}'.format(query, e)) from e
    @staticmethod
    def from_view_def(view_def):
        return one_statement_from_string(view_def + literal(sym.TERMINATOR))

class RelRef(RelExpr):
    def __init__(self, rel):
        super(RelRef, self).__init__()
        assert isinstance(rel, str)
        self.rel = rel
    def __str__(self):
        return self.rel
    def validateSubtree(self, context: StatementContext):
        # first check if this is a table in dbms:
        if context.db.table_exists(self.rel):
            attrspecs = [AttrSpec(self.rel, attr, type)\
                         for attr, type in context.db.describe(self.rel)]
            self.type = RelType(context.new_tmp(), attrspecs)
            self.view = None
            return
        # then check if this is a view defined in this session:
        view_def = context.views.raw_def(self.rel)
        if view_def is not None:
            self.view = RelExpr.from_view_def(view_def)
            self.view.validateSubtree(context)
            attrspecs = [AttrSpec(self.rel, attrspec.name, attrspec.type)\
                         for attrspec in self.view.type.attrs]
            self.type = RelType(self.view.type.tmp, attrspecs) # don't create a new temp relation
            return
        # reference is not found:
        raise self.validationError('relation {} does not exist'.format(self.rel))
    def info(self):
        if self.view is None:
            yield 'RELATION {} => {}'.format(self, self.type)
        else:
            yield 'VIEW {} => {}'.format(self, self.type)
            for i, line in enumerate(self.view.info()):
                yield ('\\_' if i == 0 else '  ') + line
    def sql(self):
        if self.view is None:
            yield '{}({}) AS (SELECT * FROM {})'\
                .format(self.type.sql_rel(), ', '.join(self.type.sql_attrs()), self.rel)
        else:
            for block in self.view.sql():
                yield block

class Rename(RelExpr):
    def __init__(self, relname, attrnames, input):
        assert relname is not None or attrnames is not None
        assert relname is None or isinstance(relname, str)
        assert attrnames is None or\
            (isinstance(attrnames, list) and len(attrnames) > 0 and\
             all(isinstance(name, str) for name in attrnames))
        assert isinstance(input, RelExpr)
        super(Rename, self).__init__([input])
        self.relname = relname
        self.attrnames = attrnames
    def __str__(self):
        s = literal(sym.RENAME) + literal(sym.ARG_L)
        if self.relname is not None:
            s += self.relname + literal(sym.COLON) + ' '
        if self.attrnames is None:
            s += literal(sym.STAR)
        else:
            s += ', '.join(name for name in self.attrnames)
        s += literal(sym.ARG_R) + ' ' + str(paren(self.inputs[0]))
        return s
    def validateSubtree(self, context: StatementContext):
        self.inputs[0].validateSubtree(context)
        if self.attrnames is not None:
            if len(self.attrnames) != len(self.inputs[0].type.attrs):
                raise self.validationError('renaming {} attribute(s) but the input has {}'\
                                           .format(len(self.attrnames),
                                                   len(self.inputs[0].type.attrs)))
            if len(self.attrnames) != len(set(self.attrnames)):
                logger.warning('{}: duplicate attribute names when renaming: {}'\
                               .format(self, ', '.join(self.attrnames)))
        else:
            for a1, a2 in itertools.combinations(self.inputs[0].type.attrs, 2):
                if a1.name == a2.name and not a1.can_be_confused_with(a2):
                    logger.warning('{}: attributes {} and {} become confused after renaming'\
                                   .format(self, a1.str_ref_only(), a2.str_ref_only()))
        attrspecs = list()
        if self.attrnames is not None:
            for name, attrspec in zip(self.attrnames, self.inputs[0].type.attrs):
                attrspecs.append(AttrSpec(self.relname, name, attrspec.type))
        else:
            for attrspec in self.inputs[0].type.attrs:
                attrspecs.append(AttrSpec(self.relname, attrspec.name, attrspec.type))
        self.type = RelType(self.inputs[0].type.tmp, attrspecs) # don't create a new temp relation
    def info(self):
        yield '{} => {}'.format(symbolic(sym.RENAME), self.type)
        for i, line in enumerate(self.inputs[0].info()):
            yield ('\\_' if i == 0 else '  ') + line
    def sql(self):
        for block in self.inputs[0].sql(): # don't create a new temp relation
            yield block

class Project(RelExpr):
    def __init__(self, attrs, input):
        assert isinstance(attrs, list)
        assert all(isinstance(attr, ValExpr) for attr in attrs)
        assert isinstance(input, RelExpr)
        super(Project, self).__init__([input])
        self.attrs = attrs
    def __str__(self):
        return literal(sym.PROJECT) + literal(sym.ARG_L) +\
            ', '.join(str(attr) for attr in self.attrs) +\
            literal(sym.ARG_R) + ' ' + str(paren(self.inputs[0]))
    def validateSubtree(self, context: StatementContext):
        self.inputs[0].validateSubtree(context)
        output_attrspecs = list()
        for attr in self.attrs:
            attr.validateSubtree(context, self)
            if isinstance(attr, AttrRef):
                _, aidx = attr.internal_ref
                output_attrspecs.append(self.inputs[0].type.attrs[aidx])
            else:
                output_attrspecs.append(AttrSpec(None, None, attr.type))
        self.type = RelType(context.new_tmp(), output_attrspecs)
    def info(self):
        yield '{} => {}'.format(symbolic(sym.PROJECT), self.type)
        for attr in self.attrs:
            yield '|' + attr.info()
        for i, line in enumerate(self.inputs[0].info()):
            yield ('\\_' if i == 0 else '  ') + line
    def sql(self):
        for block in self.inputs[0].sql():
            yield block
        yield '{}({}) AS (SELECT DISTINCT {} FROM {})'\
            .format(self.type.sql_rel(), ', '.join(self.type.sql_attrs()),
                    ', '.join(attr.sql(self) for attr in self.attrs),
                    self.inputs[0].type.sql_rel())

class Select(RelExpr):
    def __init__(self, cond, input):
        assert isinstance(cond, ValExpr)
        assert isinstance(input, RelExpr)
        super(Select, self).__init__([input])
        self.cond = cond
    def __str__(self):
        return literal(sym.SELECT) + literal(sym.ARG_L) + str(self.cond) + literal(sym.ARG_R) +\
            ' ' + str(paren(self.inputs[0]))
    def validateSubtree(self, context: StatementContext):
        self.inputs[0].validateSubtree(context)
        self.cond.validateSubtree(context, self)
        if self.cond.type != ValType.BOOLEAN:
            raise self.validationError('selection condition {} has type {}; boolean expected'\
                                       .format(self.cond, self.cond.type.value))
        self.type = RelType(context.new_tmp(), self.inputs[0].type.attrs)
    def info(self):
        yield '{} => {}'.format(symbolic(sym.SELECT), self.type)
        yield '|' + self.cond.info()
        for i, line in enumerate(self.inputs[0].info()):
            yield ('\\_' if i == 0 else '  ') + line
    def sql(self):
        for block in self.inputs[0].sql():
            yield block
        yield '{}({}) AS (SELECT * FROM {} WHERE {})'\
            .format(self.type.sql_rel(), ', '.join(self.type.sql_attrs()),
                    self.inputs[0].type.sql_rel(),
                    self.cond.sql(self))

class Join(RelExpr):
    def __init__(self, left, cond, right):
        assert isinstance(left, RelExpr)
        assert cond is None or isinstance(cond, ValExpr)
        assert isinstance(right, RelExpr)
        super(Join, self).__init__([left, right])
        self.cond = cond
    def __str__(self):
        cond = '' if self.cond is None else\
               literal(sym.ARG_L) + str(self.cond) + literal(sym.ARG_R)
        return '{} {}{} {}'.format(paren(self.inputs[0]),
                                   literal(sym.JOIN), cond,
                                   paren(self.inputs[1]))
    def validateSubtree(self, context: StatementContext):
        self.inputs[0].validateSubtree(context)
        self.inputs[1].validateSubtree(context)
        if self.cond is None:
            self.pairs = list()
            for i0, a0 in enumerate(self.inputs[0].type.attrs):
                if a0.name is None:
                    continue
                matches = [(i1, a1) for i1, a1 in enumerate(self.inputs[1].type.attrs)\
                           if a0.name == a1.name]
                if len(matches) > 1:
                    raise self.validationError('ambiguity in natural join: {} from the left input'
                                               ' matches multiple attributes on the right'\
                                               .format(a0.str_ref_only()))

                elif len(matches) == 1:
                    i1, a1 = matches[0]
                    try:
                        context.check.function_call(symbolic(sym.EQ), [a0.type, a1.type])
                    except TypeSysError as e:
                        raise self.validationError('natural join cannot equate {} and {}'\
                                                   .format(a0, a1))
                    self.pairs.append((i0, i1))
            if len(self.pairs) == 0:
                logger.warning('{}: no attributes with matching names found;'
                               ' natural join degnerates into cross product'.format(self))
            for i0, i1 in self.pairs:
                if any((j0, j1) for j0, j1 in self.pairs if j0 != i0 and j1 == i1):
                    raise self.validationError('ambiguity in natural join: {} from the right input'
                                               ' matches multiple attributes on the left'\
                                               .format(self.inputs[1].type.attrs[i1]\
                                                       .str_ref_only()))
            attrspecs = list()
            for a0 in self.inputs[0].type.attrs:
                attrspecs.append(a0)
            for i1, a1 in enumerate(self.inputs[1].type.attrs):
                if not any(i1 == i for _, i in self.pairs):
                    attrspecs.append(a1)
                    if any(a1.can_be_confused_with(a0) for a0 in self.inputs[0].type.attrs):
                        # this shouldn't happen under natural join rule, but oh well:
                        logger.warning('{}: attribute {} becomes confused with others'
                                       ' in the join output'\
                                       .format(self, a1.str_ref_only()))
            self.type = RelType(context.new_tmp(), attrspecs)
        else:
            self.cond.validateSubtree(context, self)
            if self.cond.type != ValType.BOOLEAN:
                raise self.validationError('join condition {} has type {}; boolean expected'\
                                           .format(self.cond, self.cond.type.value))
            for a0, a1 in itertools.product(self.inputs[0].type.attrs, self.inputs[1].type.attrs):
                if a0.can_be_confused_with(a1):
                    logger.warning('{}: attributes {} from the left input and {} from the right'
                                   ' become confused in the join output'\
                                   .format(self, a0.str_ref_only(), a1.str_ref_only()))
            self.type = RelType(context.new_tmp(),
                                self.inputs[0].type.attrs + self.inputs[1].type.attrs)
    def info(self):
        yield '{} => {}'.format(symbolic(sym.JOIN), self.type)
        if self.cond is not None:
            yield '|' + self.cond.info()
        elif len(self.pairs) == 0:
            yield '|inferred: cross product with no join condition'
        else:
            for i0, i1 in self.pairs:
                yield '|inferred: {}[0.{}] = {}[1.{}]'\
                    .format(self.inputs[0].type.attrs[i0].str_ref_only(), i0,
                            self.inputs[1].type.attrs[i1].str_ref_only(), i1)
        for i, line in enumerate(self.inputs[0].info()):
            yield ('\\_' if i == 0 else '| ') + line
        for i, line in enumerate(self.inputs[1].info()):
            yield ('\\_' if i == 0 else '  ') + line
    def sql(self):
        for block in self.inputs[0].sql():
            yield block
        for block in self.inputs[1].sql():
            yield block
        select = '*'
        where = ''
        if self.cond is None:
            if len(self.pairs) > 0:
                attrs = ['{}.{}'.format(self.inputs[0].type.sql_rel(), attr)\
                          for attr in self.inputs[0].type.sql_attrs()]
                attrs += ['{}.{}'.format(self.inputs[1].type.sql_rel(), attr)\
                           for i, attr in enumerate(self.inputs[1].type.sql_attrs())\
                           if all(i != i1 for _, i1 in self.pairs)]
                select = ', '.join(attrs)
                eqs = ['{}.{} = {}.{}'.format(self.inputs[0].type.sql_rel(),
                                              self.inputs[0].type.sql_attr(i0),
                                              self.inputs[1].type.sql_rel(),
                                              self.inputs[1].type.sql_attr(i1))\
                       for i0, i1 in self.pairs]
                where = ' WHERE {}'.format(' AND '.join(eqs))
        else:
            where = ' WHERE {}'.format(self.cond.sql(self))
        yield '{}({}) AS (SELECT {} FROM {}, {}{})'\
            .format(self.type.sql_rel(), ', '.join(self.type.sql_attrs()),
                    select,
                    self.inputs[0].type.sql_rel(), self.inputs[1].type.sql_rel(),
                    where)

class Cross(RelExpr):
    def __init__(self, left, right):
        assert isinstance(left, RelExpr)
        assert isinstance(right, RelExpr)
        super(Cross, self).__init__([left, right])
    def __str__(self):
        return '{} {} {}'.format(paren(self.inputs[0]), literal(sym.CROSS), paren(self.inputs[1]))
    def validateSubtree(self, context: StatementContext):
        self.inputs[0].validateSubtree(context)
        self.inputs[1].validateSubtree(context)
        for a0, a1 in itertools.product(self.inputs[0].type.attrs, self.inputs[1].type.attrs):
            if a0.can_be_confused_with(a1):
                logger.warning('{}: attributes {} from the left input and {} from the right'
                               ' become confused in the cross product output'\
                               .format(self, a0.str_ref_only(), a1.str_ref_only()))
        self.type = RelType(context.new_tmp(),
                            self.inputs[0].type.attrs + self.inputs[1].type.attrs)
    def info(self):
        yield '{} => {}'.format(symbolic(sym.CROSS), self.type)
        for i, line in enumerate(self.inputs[0].info()):
            yield ('\\_' if i == 0 else '| ') + line
        for i, line in enumerate(self.inputs[1].info()):
            yield ('\\_' if i == 0 else '  ') + line
    def sql(self):
        for block in self.inputs[0].sql():
            yield block
        for block in self.inputs[1].sql():
            yield block
        yield '{}({}) AS (SELECT * FROM {}, {})'\
            .format(self.type.sql_rel(), ', '.join(self.type.sql_attrs()),
                    self.inputs[0].type.sql_rel(), self.inputs[1].type.sql_rel())

class SetOp(RelExpr):
    def __init__(self, left, right):
        assert isinstance(left, RelExpr)
        assert isinstance(right, RelExpr)
        super(SetOp, self).__init__([left, right])
    def op(self):
        if isinstance(self, Union):
            return sym.UNION
        elif isinstance(self, Diff):
            return sym.DIFF
        elif isinstance(self, Intersect):
            return sym.INTERSECT
        else:
            assert False
    def sql_op(self):
        if isinstance(self, Union):
            return 'UNION'
        elif isinstance(self, Diff):
            return 'EXCEPT'
        elif isinstance(self, Intersect):
            return 'INTERSECT'
        else:
            assert False
    def __str__(self):
        return '{} {} {}'.format(paren(self.inputs[0]),
                                 literal(self.op()),
                                 paren(self.inputs[1]))
    def validateSubtree(self, context: StatementContext):
        self.inputs[0].validateSubtree(context)
        self.inputs[1].validateSubtree(context)
        if len(self.inputs[0].type.attrs) != len(self.inputs[1].type.attrs):
            raise self.validationError('input relations to a set operation'
                                       ' do not have the same number of attributes')
        for i, (a0, a1) in enumerate(zip(self.inputs[0].type.attrs, self.inputs[1].type.attrs)):
            if a0.type != a1.type:
                raise self.validationError('input attributes at position {} have'
                                           ' different types: {} vs. {}'.format(i, a0, a1))
            if a0.name != a1.name:
                logger.warning('{}: input attributes at position {} have'
                               ' different names: {} vs. {}'.format(self, i,
                                                                    a0.str_ref_only(),
                                                                    a1.str_ref_only()))
        self.type = RelType(context.new_tmp(), self.inputs[0].type.attrs)
    def info(self):
        yield '{} => {}'.format(symbolic(self.op()), self.type)
        for i, line in enumerate(self.inputs[0].info()):
            yield ('\\_' if i == 0 else '| ') + line
        for i, line in enumerate(self.inputs[1].info()):
            yield ('\\_' if i == 0 else '  ') + line
    def sql(self):
        for block in self.inputs[0].sql():
            yield block
        for block in self.inputs[1].sql():
            yield block
        # interestingly, in the SQL below, parentheses around the two
        # SELECT subqueries are not needed, and SQLite in fact would
        # not like them:
        yield '{}({}) AS (SELECT * FROM {} {} SELECT * FROM {})'\
            .format(self.type.sql_rel(), ', '.join(self.type.sql_attrs()),
                    self.inputs[0].type.sql_rel(),
                    self.sql_op(),
                    self.inputs[1].type.sql_rel())

class Union(SetOp):
    pass

class Diff(SetOp):
    pass

class Intersect(SetOp):
    pass

class Define(Node):
    def __init__(self, view, definition):
        assert isinstance(view, str)
        assert isinstance(definition, RelExpr)
        self.view = view
        self.definition = definition
    def __str__(self):
        return '{} {} {}'.format(self.view, literal(sym.GETS), self.definition)
    def validate(self, context):
        if context.db.table_exists(self.view):
            raise ValidationError('{} is a database relation'.format(self.view), self)
        if context.views.raw_def(self.view) is None:
            self.definition.validate(context)
        else:
            dependents = context.views.find_dependents(self.view, recurse=True)
            # validate this definition without previously dependent views:
            tmp_context = Context(context.db, context.check, context.views.clone())
            tmp_context.views.clear(self.view) # which removes dependents too
            try:
                self.definition.validate(tmp_context)
            except ValidationError as e:
                if isinstance(e.node, RelRef) and\
                   (e.node.rel in dependents or e.node.rel == self.view):
                    raise ValidationError('{} and {} are circularly defined'\
                                          .format(self.view, e.node.rel), self)\
                                          from e
                else:
                    raise
            # now, with the new definition, try validating dependents:
            tmp_context.views = context.views.clone()
            tmp_context.views.register(self.view,
                                       str(self.definition), self.definition.find_views())
            for v in dependents:
                try:
                    ast = RelExpr.from_view_def(tmp_context.views.raw_def(v))
                    ast.validate(tmp_context)
                except ValidationError as e:
                    raise ValidationError('view definition of {} is broken'
                                          ' by the new defintion of {}:\n{}'\
                                          .format(v, self.view, e), self)\
                                          from e
    def info(self):
        views = self.definition.find_views()
        yield 'VIEW {} depends on '.format(self.view) +\
            ('no other views' if len(views) == 0 else ', '.join(views))
        yield literal(sym.GETS)
        for line in self.definition.info():
            yield line
    def execute(self, context):
        context.views.register(self.view, str(self.definition), self.definition.find_views())
        print('view {} defined'.format(self.view))

class Command(Node):
    def __init__(self, cmd):
        assert cmd in (sym.LIST, sym.QUIT)
        self.cmd = cmd
    def __str__(self):
        return literal(self.cmd)
    def validate(self, context):
        pass
    def info(self):
        yield symbolic(self.cmd)
    def execute(self, context):
        if self.cmd == sym.LIST:
            print('database relations:')
            for table in context.db.list():
                attrs = ['{}:{}'.format(attr, type.value)\
                         for attr, type in context.db.describe(table)]
                print('  {}({})'.format(table, ', '.join(attrs)))
            if len(context.views.list()) > 0:
                print('views defined:')
                for view in context.views.list():
                    raw_def = context.views.raw_def(view)
                    ast = RelExpr.from_view_def(raw_def)
                    logging.disable(logging.WARNING)
                    ast.validate(context)
                    logging.disable(logging.NOTSET)
                    print('  {}({}) {} {}'.format(view,
                                                  ', '.join(ast.type.str_attr_names_and_types()),
                                                  literal(sym.GETS), raw_def))
        elif self.cmd == sym.QUIT:
            sys.exit(0)

class CommandClear(Command):
    def __init__(self, force=False, view=None):
        assert isinstance(force, bool)
        self.force = force
        assert view is None or isinstance(view, str)
        self.view = view
    def __str__(self):
        return literal(sym.CLEAR) +\
            (literal(sym.FORCE) if self.force else '') + ' ' +\
            (literal(sym.STAR) if self.view is None else self.view)
    def validate(self, context):
        if self.view is not None:
            if context.views.raw_def(self.view) is None:
                raise ValidationError('view {} is not defined'.format(self.view), self)
            if not self.force:
                dependents = context.views.find_dependents(self.view)
                if len(dependents) > 0:
                    raise ValidationError('view {} is used in defining others ({});'
                                          ' to clear them all, use {} after {}'\
                                          .format(self.view, ', '.join(dependents),
                                                  literal(sym.FORCE), literal(sym.CLEAR)), self)
    def info(self):
        yield lreplace(str(self), literal(sym.CLEAR), symbolic(sym.CLEAR))
    def execute(self, context):
        views_cleared = context.views.list() if self.view is None else\
                        context.views.find_dependents(self.view) | {self.view}
        if len(views_cleared) == 0:
            print('no views to clear')
        else:
            context.views.clear(self.view)
            print('views cleared: {}'.format(', '.join(views_cleared)))

class CommandSave(Command):
    def __init__(self, force=False, view=None, filename=None):
        assert isinstance(force, bool)
        self.force = force
        assert view is None or isinstance(view, str)
        self.view = view
        assert filename is None or isinstance(filename, str)
        if filename is None:
            self.filename = '{}.ra'.format('views' if self.view is None else self.view)
        else:
            self.filename = sqlstr_to_str(filename)
    def __str__(self):
        return literal(sym.SAVE) +\
            (literal(sym.FORCE) if self.force else '') + ' ' +\
            (literal(sym.STAR) if self.view is None else self.view) + ' ' +\
            str_to_sqlstr(self.filename)
    def validate(self, context):
        if self.view is not None and context.views.raw_def(self.view) is None:
            raise ValidationError('view {} is not defined'.format(self.view), self)
        if not self.force and os.path.isfile(self.filename):
            raise ValidationError('file {} already exists; to overwrite use {} after {}'\
                                  .format(self.filename, literal(sym.FORCE), literal(sym.SAVE)),
                                  self)
    def info(self):
        yield lreplace(str(self), literal(sym.SAVE), symbolic(sym.SAVE))
    def execute(self, context):
        views = context.views.topo(self.view)
        try:
            with open(self.filename, 'w') as f:
                for v in views:
                    print('{} {} {};'.format(v, literal(sym.GETS), context.views.raw_def(v)),
                          file=f)
            print('wrote file {}'.format(self.filename))
        except Exception as e:
            raise ExecutionError('writing to {} failed'.format(self.filename), self) from e

class CommandSource(Command):
    def __init__(self, filename):
        assert isinstance(filename, str)
        self.filename = sqlstr_to_str(filename)
    def __str__(self):
        return literal(sym.SOURCE) + ' ' + str_to_sqlstr(self.filename)
    def validate(self, context):
        if not os.path.isfile(self.filename):
            raise ValidationError('file {} does not exist'.format(self.filename), self)
    def info(self):
        yield lreplace(str(self), literal(sym.SOURCE), symbolic(sym.SOURCE))
    def execute(self, context):
        try:
            execute_from_file(self.filename, context, echo=True)
        except Exception as e:
            raise ExecutionError('error while executing from {}:\n{}'\
                                 .format(self.filename, e)) from e

class CommandSqlexec(Command):
    def __init__(self, sql):
        assert isinstance(sql, str)
        self.sql = sql
    def __str__(self):
        return literal(sym.SQLEXEC) +\
            literal(sym.ARG_L) + self.sql + literal(sym.ARG_R)
    def validate(self, context):
        pass
    def info(self):
        yield symbolic(sym.SQLEXEC) + ': ' + self.sql
    def execute(self, context):
        logger.debug(self.sql)
        try:
            context.db.execute_and_print_result(self.sql)
        except Exception as e:
            raise ExecutionError('SQL error in:\n{}\n{}'.format(self.sql, e)) from e



import enum
import sys
import importlib
import json

def lreplace(s, prefix, sub):
    return s.replace(prefix, sub, 1)

def rreplace(s, suffix, sub):
    return sub.join(s.rsplit(suffix, 1))

def AutoNameEnum(enum_type_name, enum_fields):
    return enum.Enum(enum_type_name, [(field, field.lower()) for field in enum_fields])

class Tee:
    def __init__(self, filename):
        self.terminal = sys.stdout
        self.file = open(filename, 'w')

    def write(self, s):
        self.terminal.write(s)
        self.file.write(s)

    def flush(self):
        self.terminal.flush()
        self.file.flush()

def number_lines(s):
    lines = s.split('\n')
    lineno_max_width = len(str(len(lines)))
    for i, line in enumerate(lines):
        yield ('line {:>' + str(lineno_max_width) + '}: {}').format(i+1, line)

def dynamic_import(full_class_name):
    parts = full_class_name.split('.')
    mod = __import__('.'.join(parts[0:-1]), fromlist=[parts[-1]])
    return getattr(mod, parts[-1])

class CustomJSONSerializable:
    def to_json(self):
        return json.dumps(self, cls=CustomJSONEncoder, indent=2, sort_keys=True)
    def list_to_json(collection):
        return json.dumps(collection, cls=CustomJSONEncoder, indent=2, sort_keys=True)
    def from_json(object_json):
        if isinstance(object_json, str):
            return json.loads(object_json, cls=CustomJSCONDecoder)
        else:
            return json.load(object_json, cls=CustomJSCONDecoder)
    def list_from_json(list_json):
        if isinstance(list_json, str):
            return json.loads(list_json, cls=CustomJSCONDecoder)
        else:
            return json.load(list_json, cls=CustomJSCONDecoder)

class CustomJSONEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, dict):
            return {k : self.default(v) for k, v in obj.items()}
        elif isinstance(obj, list) and not isinstance(obj, str):
            return [self.default(v) for v in obj]
        elif isinstance(obj, CustomJSONSerializable):
            outdict = {a : self.default(v) for a, v in obj.__dict__.items()}
            outdict['__type__'] = obj.__class__.__module__ + '.' + obj.__class__.__qualname__
            return outdict
        return obj

class CustomJSCONDecoder(json.JSONDecoder):
    def __init__(self, *args, **kwargs):
        json.JSONDecoder.__init__(self, object_hook=self.object_hook, *args, **kwargs)
    def object_hook(self, o):
        try:
            if '__type__' in o:
                cls = dynamic_import(o['__type__'])
                obj = cls(**{key : o[key] for key in o if key != '__type__'})
                return obj
        except:
            pass
        return o



# Generated from RAParser.g4 by ANTLR 4.7
from antlr4 import *
if __name__ is not None and "." in __name__:
    from .RAParser import RAParser
else:
    from RAParser import RAParser

# This class defines a complete generic visitor for a parse tree produced by RAParser.

class RAParserVisitor(ParseTreeVisitor):

    # Visit a parse tree produced by RAParser#andExpr.
    def visitAndExpr(self, ctx:RAParser.AndExprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by RAParser#numberLiteralValExpr.
    def visitNumberLiteralValExpr(self, ctx:RAParser.NumberLiteralValExprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by RAParser#likeExpr.
    def visitLikeExpr(self, ctx:RAParser.LikeExprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by RAParser#funcExpr.
    def visitFuncExpr(self, ctx:RAParser.FuncExprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by RAParser#plusMinusExpr.
    def visitPlusMinusExpr(self, ctx:RAParser.PlusMinusExprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by RAParser#compareExpr.
    def visitCompareExpr(self, ctx:RAParser.CompareExprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by RAParser#orExpr.
    def visitOrExpr(self, ctx:RAParser.OrExprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by RAParser#attrRef.
    def visitAttrRef(self, ctx:RAParser.AttrRefContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by RAParser#multDivExpr.
    def visitMultDivExpr(self, ctx:RAParser.MultDivExprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by RAParser#concatExpr.
    def visitConcatExpr(self, ctx:RAParser.ConcatExprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by RAParser#valExprParenthesized.
    def visitValExprParenthesized(self, ctx:RAParser.ValExprParenthesizedContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by RAParser#isNotNullExpr.
    def visitIsNotNullExpr(self, ctx:RAParser.IsNotNullExprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by RAParser#isNullExpr.
    def visitIsNullExpr(self, ctx:RAParser.IsNullExprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by RAParser#stringLiteralValExpr.
    def visitStringLiteralValExpr(self, ctx:RAParser.StringLiteralValExprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by RAParser#notExpr.
    def visitNotExpr(self, ctx:RAParser.NotExprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by RAParser#listOfValExprs.
    def visitListOfValExprs(self, ctx:RAParser.ListOfValExprsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by RAParser#listOfIDs.
    def visitListOfIDs(self, ctx:RAParser.ListOfIDsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by RAParser#crossExpr.
    def visitCrossExpr(self, ctx:RAParser.CrossExprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by RAParser#relExprParenthesized.
    def visitRelExprParenthesized(self, ctx:RAParser.RelExprParenthesizedContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by RAParser#projectExpr.
    def visitProjectExpr(self, ctx:RAParser.ProjectExprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by RAParser#relRef.
    def visitRelRef(self, ctx:RAParser.RelRefContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by RAParser#intersectExpr.
    def visitIntersectExpr(self, ctx:RAParser.IntersectExprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by RAParser#selectExpr.
    def visitSelectExpr(self, ctx:RAParser.SelectExprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by RAParser#diffExpr.
    def visitDiffExpr(self, ctx:RAParser.DiffExprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by RAParser#joinExpr.
    def visitJoinExpr(self, ctx:RAParser.JoinExprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by RAParser#renameExpr.
    def visitRenameExpr(self, ctx:RAParser.RenameExprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by RAParser#unionExpr.
    def visitUnionExpr(self, ctx:RAParser.UnionExprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by RAParser#definition.
    def visitDefinition(self, ctx:RAParser.DefinitionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by RAParser#listCommand.
    def visitListCommand(self, ctx:RAParser.ListCommandContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by RAParser#clearCommand.
    def visitClearCommand(self, ctx:RAParser.ClearCommandContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by RAParser#saveCommand.
    def visitSaveCommand(self, ctx:RAParser.SaveCommandContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by RAParser#sourceCommand.
    def visitSourceCommand(self, ctx:RAParser.SourceCommandContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by RAParser#quitCommand.
    def visitQuitCommand(self, ctx:RAParser.QuitCommandContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by RAParser#sqlexecCommand.
    def visitSqlexecCommand(self, ctx:RAParser.SqlexecCommandContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by RAParser#statement.
    def visitStatement(self, ctx:RAParser.StatementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by RAParser#program.
    def visitProgram(self, ctx:RAParser.ProgramContext):
        return self.visitChildren(ctx)



del RAParser


import subprocess
import logging
import json
import importlib

from sqlalchemy import engine, create_engine, inspect, text
import sqlalchemy.types
import sqlalchemy.exc

from radb.typesys import ValType

def sqltype_to_ratype(sqltype):
    if isinstance(sqltype, sqlalchemy.types.Boolean):
        return ValType.BOOLEAN
    elif isinstance(sqltype, sqlalchemy.types.Integer) or\
         isinstance(sqltype, sqlalchemy.types.Numeric):
        return ValType.NUMBER
    elif isinstance(sqltype, sqlalchemy.types.String):
        return ValType.STRING
    elif isinstance(sqltype, sqlalchemy.types.Date):
        return ValType.DATE
    elif isinstance(sqltype, sqlalchemy.types.DateTime):
        return ValType.DATETIME
    else:
        return ValType.UNKNOWN

class DB:

    def __init__(self, configured, prefix='db.'):
        props = { key[len(prefix):] : configured[key]\
                  for key in configured if key.startswith(prefix) }
        self.engine = create_engine(engine.url.URL(**props))
        self.inspector = inspect(self.engine)
        self.conn = self.engine.connect()

    def list(self):
        return self.inspector.get_table_names()

    def table_exists(self, table):
        # safer than comparing against self.list() because it seems to
        # respect the case-sensivitity convention of the dbms.
        return self.engine.dialect.has_table(self.conn, table)

    def describe(self, table):
        attrs = list()
        for d in self.inspector.get_columns(table):
            attrs.append((d['name'], sqltype_to_ratype(d['type'])))
        return attrs

    def execute(self, query, **kwargs):
        try:
            result = self.conn.execute(text(query), **kwargs)
        except sqlalchemy.exc.DatabaseError:
            conn = self.engine.connect()
            result = self.conn.execute(text(query), **kwargs)
        return result

    def execute_and_print_result(self, query, **kwargs):
        result = self.execute(query, **kwargs)
        if result.returns_rows:
            print('-'*70)
            count = 0
            for row in result:
                print(', '.join(str(val) for val in row))
                count += 1
            print('-'*70)
            print('{} tuple{} returned'.format('no' if count == 0 else count,
                                               '' if count == 1 else 's'))
        else:
            print('done')



from setuptools import setup, find_packages
from codecs import open
import os
import configparser

def get_path(fname):
    return os.path.join(os.path.abspath(os.path.dirname(__file__)), fname)

def parse_setup_config(fname):
    configfile = get_path(fname)
    config = configparser.ConfigParser()
    config.read(configfile)
    return { key.replace('setup.', '', 1): val
             for key, val in config.items(configparser.DEFAULTSECT)
             if key.startswith('setup.') }
setup_config = parse_setup_config('radb/sys.ini')

def read(fname):
    return open(get_path(fname), encoding='utf-8').read()

setup(
    **setup_config,
    long_description=read('README.rst'),
    license='MIT',
    keywords='database relational algebra',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Education',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Topic :: Database',
        'Topic :: Database :: Front-Ends',
        'Topic :: Software Development :: Interpreters',
        'Programming Language :: Python :: 3',
    ],

    packages=find_packages(),
    install_requires=['sqlalchemy', 'antlr4-python3-runtime'],
    python_requires='>=3.5',
    include_package_data=True,
    entry_points={
        'console_scripts': [
            'radb=radb.ra:main',
        ],
    },

)
